from geoalchemy2.shape import from_shape, to_shape
from marshmallow import Schema, fields, pre_dump, validates
from shapely.geometry import box

from ..models import (
    AttributeType,
    BaseData,
    BaseGeoData,
    DataAttribute,
    DataAttributeNominal,
    DataAttributeOrdinal,
    DataAttributeQuantitative,
    Feature,
    FeatureType,
    GeneratedGeoData,
    GeoData,
    GeoDataType,
    GlobalData,
    KeepOverlap,
    ModelType,
    ProjectData,
    WFSGeoData,
    WMSGeoData,
)
from .base import (
    BaseSchema,
    FieldMethod,
    FieldPluck,
    OneOfSchema,
    constant_enum,
)
from .schemas import UserSchema
from .utils import validate_name


class BaseDataAttributeSchema(BaseSchema):
    class Meta(BaseSchema.Meta):
        model = DataAttribute
        fields = ("id", "name", "statistics", "data", "type")
        # This schema is never use for creation nor update
        # so basically all fields are not updatable nor creatable (no write)
        dump_only = ("id", "name", "statistics", "data", "type")

    data = fields.Nested("BaseDataSchema", only=("id", "name"))
    type = fields.Enum(
        AttributeType,
        required=True,
        metadata={"description": "Type of attribute"},
    )

    def get_values(self, obj):
        return [v.value for v in obj.values]


class BaseDataAttributeSchemaWithValues(BaseDataAttributeSchema):
    class Meta(BaseDataAttributeSchema.Meta):
        model = DataAttribute
        fields = ("id", "name", "statistics", "data", "values", "type")
        # This schema is never use for creation nor update
        # so basically all fields are not updatable nor creatable (no write)
        dump_only = ("id", "name", "statistics", "data", "values", "type")

    values = FieldMethod(fields.List(fields.Dict), "get_values")

    def get_values(self, obj):
        return [v.value for v in obj.values]


class DataAttributeQuantitativeSchema(BaseDataAttributeSchema):
    class Meta(BaseDataAttributeSchema.Meta):
        model = DataAttributeQuantitative

    type = fields.Enum(
        constant_enum(AttributeType.quantitative), required=True
    )


class DataAttributeNominalSchema(BaseDataAttributeSchema):
    class Meta(BaseDataAttributeSchema.Meta):
        model = DataAttributeNominal

    type = fields.Enum(constant_enum(AttributeType.nominal), required=True)


class DataAttributeOrdinalSchema(BaseDataAttributeSchema):
    class Meta(BaseDataAttributeSchema.Meta):
        model = DataAttributeOrdinal
        fields = ("id", "name", "statistics", "order", "data", "type")
        # This schema is never use for creation nor update
        # so basically all fields are not updatable nor creatable (no write)
        dump_only = ("id", "name", "statistics", "order", "data", "type")

    type = fields.Enum(constant_enum(AttributeType.ordinal), required=True)


class DataAttributeQuantitativeSchemaWithValues(
    BaseDataAttributeSchemaWithValues
):
    class Meta(BaseDataAttributeSchemaWithValues.Meta):
        model = DataAttributeQuantitative

    type = fields.Enum(
        constant_enum(AttributeType.quantitative), required=True
    )


class DataAttributeNominalSchemaWithValues(BaseDataAttributeSchemaWithValues):
    class Meta(BaseDataAttributeSchemaWithValues.Meta):
        model = DataAttributeNominal

    type = fields.Enum(constant_enum(AttributeType.nominal), required=True)


class DataAttributeOrdinalSchemaWithValues(BaseDataAttributeSchemaWithValues):
    class Meta(BaseDataAttributeSchemaWithValues.Meta):
        model = DataAttributeOrdinal
        fields = (
            "id",
            "name",
            "statistics",
            "order",
            "data",
            "values",
            "type",
        )
        # This schema is never use for creation nor update
        # so basically all fields are not updatable nor creatable (no write)
        dump_only = (
            "id",
            "name",
            "statistics",
            "order",
            "data",
            "values",
            "type",
        )

    type = fields.Enum(constant_enum(AttributeType.ordinal), required=True)


class DataAttributeSchema(OneOfSchema, BaseDataAttributeSchema):
    type_schemas = {
        AttributeType.quantitative.name: DataAttributeQuantitativeSchema,
        AttributeType.nominal.name: DataAttributeNominalSchema,
        AttributeType.ordinal.name: DataAttributeOrdinalSchema,
    }

    def get_obj_type(self, obj):
        return obj.type.name


class DataAttributeSchemaWithValues(
    OneOfSchema, BaseDataAttributeSchemaWithValues
):
    type_schemas = {
        AttributeType.quantitative.name: (
            DataAttributeQuantitativeSchemaWithValues
        ),
        AttributeType.nominal.name: DataAttributeNominalSchemaWithValues,
        AttributeType.ordinal.name: DataAttributeOrdinalSchemaWithValues,
    }

    def get_obj_type(self, obj):
        return obj.type.name


class BaseDataSchema(BaseSchema):
    class Meta(BaseSchema.Meta):
        model = BaseData
        fields = (
            "id",
            "type",
            "name",
            "description",
            "attributes",
            "created_at",
            "modified_at",
        )
        # This schema is never use for creation, so all below fields are the
        # non-updatable
        # is_public can only be changed through permission schema
        dump_only = ("id", "type", "attributes", "created_at", "modified_at")

    type = fields.Enum(GeoDataType)
    name = fields.String(required=False)
    attributes = fields.Nested(
        DataAttributeSchema, many=True, exclude=("data",)
    )

    @validates("name")
    def validate_name(self, value, **kwargs):
        validate_name(value)

    @pre_dump
    def load_properties(self, data, **kwargs):
        """Load the computed properties of the object (stats, attributes...)"""
        data.load_properties()
        return data


class BaseGeoDataSchema(BaseDataSchema):
    class Meta(BaseDataSchema.Meta):
        model = BaseGeoData
        fields = (
            "id",
            "type",
            "name",
            "description",
            "attributes",
            "extent",
            "extent_filter",
            "created_at",
            "modified_at",
        )
        # This schema is never use for creation, so all below fields are the
        # non-updatable
        # is_public can only be changed through permission schema
        dump_only = (
            "id",
            "type",
            "attributes",
            "extent",
            "extent_filter",
            "created_at",
            "modified_at",
        )

    extent = FieldMethod(
        fields.List(fields.Float()),
        "get_extent",
        metadata={"minItems": 4, "maxItems": 4},
    )
    extent_filter = FieldMethod(
        fields.List(fields.Float()),
        "get_extent_filter",
        "set_extent_filter",
        metadata={"minItems": 4, "maxItems": 4},
    )

    def get_extent(self, obj):
        """Returns the extent of the data (xmin,ymin,xmax,ymax)"""
        if obj.extent is None:
            return [-180.0, -90.0, 180.0, 90.0]
        extent = to_shape(obj.extent)
        return extent.bounds

    def get_extent_filter(self, obj):
        """Returns the extent filter of the data (xmin,ymin,xmax,ymax)"""
        if obj.extent_filter is None:
            return [-180.0, -90.0, 180.0, 90.0]
        extent_filter = to_shape(obj.extent_filter)
        return extent_filter.bounds

    def set_extent_filter(self, extent):
        """Set the extent filter of the data (xmin,ymin,xmax,ymax)"""
        if extent:
            geom = box(*extent)
            return from_shape(geom)
        return None


class WFSGeoDataSchema(BaseGeoDataSchema):
    class Meta(BaseGeoDataSchema.Meta):
        model = WFSGeoData
        fields = (
            "id",
            "type",
            "name",
            "description",
            "attributes",
            "extent",
            "extent_filter",
            "created_at",
            "modified_at",
            "is_public",
            "original_name",
            "title",
            "url",
            "keywords",
            "version",
            "feature_type",
            "upload_user",
        )
        # This schema is never use for creation, so all below fields are the
        # non-updatable
        # is_public can only be changed through permission schema
        dump_only = (
            "id",
            "url",
            "type",
            "attributes",
            "extent",
            "is_public",
            "feature_type",
            "upload_user",
            "original_name",
            "created_at",
            "modified_at",
        )

    type = fields.Enum(constant_enum(GeoDataType.wfs))
    upload_user = fields.Nested(UserSchema, only=("id", "username"))


class WFSGeoDataCreationSchema(WFSGeoDataSchema):
    class Meta(WFSGeoDataSchema.Meta):
        exclude = (
            "id",
            "type",
            "name",
            "attributes",
            "extent",
            "original_name",
            "title",
            "description",
            "keywords",
            "upload_user",
            "is_public",
            "created_at",
            "modified_at",
        )
        dump_only = ()


class WMSGeoDataSchema(BaseGeoDataSchema):
    class Meta(BaseGeoDataSchema.Meta):
        model = WMSGeoData
        fields = (
            "id",
            "type",
            "name",
            "description",
            "attributes",
            "extent",
            "extent_filter",
            "created_at",
            "modified_at",
            "is_public",
            "original_name",
            "title",
            "url",
            "keywords",
            "version",
            "classes",
            "start",
            "stop",
            "step",
            "layer",
            "resolution",
            "upload_user",
        )
        dump_only = (
            "id",
            "url",
            "type",
            "attributes",
            "extent",
            "is_public",
            "layer",
            "upload_user",
            "original_name",
            "created_at",
            "modified_at",
        )

    type = fields.Enum(constant_enum(GeoDataType.wms))
    classes = fields.List(fields.List(fields.Float))
    upload_user = fields.Nested(UserSchema, only=("id", "username"))


class WMSGeoDataCreationSchema(WMSGeoDataSchema):
    class Meta(WMSGeoDataSchema.Meta):
        exclude = (
            "id",
            "type",
            "name",
            "original_name",
            "title",
            "description",
            "keywords",
            "upload_user",
            "is_public",
            "created_at",
            "modified_at",
        )
        dump_only = ()


class StreamGeoDataSchema(OneOfSchema, BaseGeoDataSchema):
    type_field = "type"
    type_schemas = {
        GeoDataType.wfs.name: WFSGeoDataSchema,
        GeoDataType.wms.name: WMSGeoDataSchema,
    }

    def get_obj_type(self, obj):
        return obj.type.name


class StreamGeoDataPermissionSchema(BaseSchema):
    class Meta(BaseSchema.Meta):
        model = BaseData.Permission
        # All fields can be used for creation, update and retrieval
        fields = ("user",)

    user = fields.Pluck(UserSchema, "id")


class StreamGeoDataAccessSchema(BaseSchema):
    class Meta(BaseSchema.Meta):
        model = BaseData
        # All fields can be used for creation, update and retrieval
        fields = ("is_public", "permissions")

    permissions = fields.Nested(StreamGeoDataPermissionSchema, many=True)


class GlobalDataSchema(BaseDataSchema):
    class Meta(BaseDataSchema.Meta):
        model = GlobalData
        fields = (
            "id",
            "type",
            "name",
            "description",
            "attributes",
            "properties",
            "upload_user",
            "is_public",
            "created_at",
            "modified_at",
        )
        # This schema is never use for creation, so all below fields are the
        # non-updatable
        # is_public can only be changed through permission schema
        dump_only = (
            "id",
            "type",
            "attributes",
            "upload_user",
            "is_public",
            "created_at",
            "modified_at",
        )

    type = fields.Enum(constant_enum(GeoDataType.global_data))
    upload_user = fields.Nested(UserSchema, only=("id", "username"))
    properties = FieldMethod(
        fields.Dict(),
        "get_properties",
        deserialize="set_properties",
        metadata={"description": "properties organized by property name"},
        required=False,
    )

    def get_properties(self, obj):
        return {p.attribute.name: p.value for p in obj.properties}

    def set_properties(self, properties):
        # If persisting changes
        #    need to delete old properties beforehand
        #    and add attribute with properties afterwards
        attributes = {attr.name: attr for attr in self.instance.attributes}
        new_properties = []
        for key, value in properties.items():
            old_attr = attributes.get(key, None)
            attr_type = self.instance._choose_attribute_type(value)
            attr = (
                old_attr
                if isinstance(old_attr, attr_type)
                else attr_type(name=key, data=self.instance)
            )
            new_properties.append(
                attr.get_value_class()(
                    value=value, attribute=attr, feature=self.instance.feature
                )
            )
        return new_properties


class GlobalDataCreationSchema(Schema):
    name = fields.String(required=True)
    description = fields.String(required=False)
    properties = fields.Dict(
        metadata={"description": "properties organized by property name"},
        required=True,
    )

    @validates("name")
    def validate_name(self, value, **kwargs):
        validate_name(value)


class GeoDataSchema(BaseGeoDataSchema):
    class Meta(BaseGeoDataSchema.Meta):
        model = GeoData
        update_fields = ("name", "description")
        fields = (
            "id",
            "type",
            "name",
            "original_name",
            "source_driver",
            "upload_user",
            "description",
            "attributes",
            "extent",
            "extent_filter",
            "created_at",
            "modified_at",
            "is_public",
        )
        # This schema is never use for creation, so all below fields are the
        # non-updatable
        # is_public can only be changed through permission schema
        dump_only = (
            "id",
            "type",
            "original_name",
            "source_driver",
            "upload_user",
            "attributes",
            "extent",
            "extent_filter",
            "created_at",
            "modified_at",
            "is_public",
        )

    type = fields.Enum(constant_enum(GeoDataType.geo_data))
    upload_user = fields.Nested(UserSchema, only=("id", "username"))


class GeneratedGeoDataSchema(BaseGeoDataSchema):
    class Meta(BaseGeoDataSchema.Meta):
        model = GeneratedGeoData
        fields = (
            "id",
            "type",
            "name",
            "attributes",
            "extent",
            "created_at",
            "modified_at",
            "project_data",
        )
        # This schema is never use for creation, so all below fields are the
        # non-updatable
        # is_public can only be changed through permission schema
        dump_only = (
            "id",
            "type",
            "attributes",
            "extent",
            "created_at",
            "modified_at",
            "project_data",
        )

    type = fields.Enum(constant_enum(GeoDataType.generated_geo_data))
    project_data = FieldPluck(ProjectData, "id")


class DataSchema(OneOfSchema, BaseDataSchema):
    type_field = "type"
    type_schemas = {
        GeoDataType.global_data.name: GlobalDataSchema,
        GeoDataType.geo_data.name: GeoDataSchema,
        GeoDataType.generated_geo_data.name: GeneratedGeoDataSchema,
        GeoDataType.wfs.name: WFSGeoDataSchema,
        GeoDataType.wms.name: WMSGeoDataSchema,
    }

    def get_obj_type(self, obj):
        return obj.type.name


class GeoOnlyDataSchema(OneOfSchema, BaseDataSchema):
    type_field = "type"
    type_schemas = {
        GeoDataType.geo_data.name: GeoDataSchema,
        GeoDataType.generated_geo_data.name: GeneratedGeoDataSchema,
        GeoDataType.wfs.name: WFSGeoDataSchema,
        GeoDataType.wms.name: WMSGeoDataSchema,
    }

    def get_obj_type(self, obj):
        return obj.type.name


class GeoDataPermissionSchema(BaseSchema):
    class Meta(BaseSchema.Meta):
        model = GeoData.Permission
        # All fields can be used for creation, update and retrieval
        fields = ("user",)

    user = fields.Pluck(UserSchema, "id")


class GeoDataAccessSchema(BaseSchema):
    class Meta(BaseSchema.Meta):
        model = GeoData
        # All fields can be used for creation, update and retrieval
        fields = ("is_public", "permissions")

    permissions = fields.Nested(GeoDataPermissionSchema, many=True)


class GlobalDataPermissionSchema(BaseSchema):
    class Meta(BaseSchema.Meta):
        model = GlobalData.Permission
        # All fields can be used for creation, update and retrieval
        fields = ("user",)

    user = fields.Pluck(UserSchema, "id")


class GlobalDataAccessSchema(BaseSchema):
    class Meta(BaseSchema.Meta):
        model = GlobalData
        # All fields can be used for creation, update and retrieval
        fields = ("is_public", "permissions")

    permissions = fields.Nested(GlobalDataPermissionSchema, many=True)


class BaseDisaggregationCriterionSchema(Schema):
    data = fields.String()
    attribute = fields.String()
    value = fields.Raw()


class DisaggregationCriterionWeightedSumSchema(
    BaseDisaggregationCriterionSchema
):
    weight = fields.Float()


class DisaggregationCriterionZonePropositionSchema(
    BaseDisaggregationCriterionSchema
):
    numeric_value = fields.Float()
    area = fields.Float()
    fitness = fields.Float()


class BaseExplainabilitySchema(Schema):
    model_type = fields.Enum(ModelType)


class BaseModelExplainabilitySchema(Schema):
    disaggregation = fields.Nested(
        BaseDisaggregationCriterionSchema, many=True
    )


class ExplainabilityContinuousRuleSchema(BaseModelExplainabilitySchema):
    model_type = fields.Enum(constant_enum(ModelType.continuous_rule))


class ExplainabilityBufferSchema(BaseModelExplainabilitySchema):
    model_type = fields.Enum(constant_enum(ModelType.geo_buffer))
    radius = fields.Integer()


class ExplainabilityWeightedSumSchema(BaseModelExplainabilitySchema):
    model_type = fields.Enum(constant_enum(ModelType.weighted_sum))
    disaggregation = fields.Nested(
        DisaggregationCriterionWeightedSumSchema, many=True
    )


class ExplainabilityDiscreteModelSchema(BaseModelExplainabilitySchema):
    model_type = fields.Enum(constant_enum(ModelType.categories_rule))
    rule = fields.String()


class ProfileCriterionSchema(Schema):
    data = fields.String(allow_none=False)
    attribute = fields.String(allow_none=False)
    value = fields.Raw(allow_none=False)
    weight = fields.Float(allow_none=False)


class ExplainabilityMRSortCriterionSchema(Schema):
    data = fields.String()
    attribute = fields.String()
    maximize = fields.Boolean()


class ExplainabilityMRSortSchema(BaseModelExplainabilitySchema):
    model_type = fields.Enum(constant_enum(ModelType.mrsort))
    lower = fields.Nested(ProfileCriterionSchema, many=True, allow_none=True)
    upper = fields.Nested(ProfileCriterionSchema, many=True, allow_none=True)
    categories = fields.List(fields.String(allow_none=False))
    criteria = fields.Nested(
        BaseDisaggregationCriterionSchema,
        only=("data", "attribute"),
        many=True,
    )
    majority_threshold = fields.Float()


class ExplainabilityAttributeWithValues(Schema):
    data = fields.String()
    attribute = fields.String()
    values = fields.List(fields.Raw(allow_none=False))


class ExplainabilityOverlapSchema(BaseExplainabilitySchema):
    model_type = fields.Enum(constant_enum(ModelType.merge_overlap))
    aggregated = fields.Nested(ExplainabilityAttributeWithValues, many=True)
    keep_overlap = fields.Enum(KeepOverlap)


class ExplainabilityMergedFeaturesSchema(BaseExplainabilitySchema):
    model_type = fields.Enum(constant_enum(ModelType.dissolve_adjacent))


class ExplainabilityZonePropositionSchema(BaseModelExplainabilitySchema):
    model_type = fields.Enum(constant_enum(ModelType.zone_proposition))
    disaggregation = fields.Nested(
        DisaggregationCriterionZonePropositionSchema, many=True
    )


class ExplainabilitySchema(OneOfSchema, BaseExplainabilitySchema):
    type_field = "model_type"
    type_schemas = {
        ModelType.categories_rule.name: ExplainabilityDiscreteModelSchema,
        ModelType.continuous_rule.name: ExplainabilityContinuousRuleSchema,
        ModelType.geo_buffer.name: ExplainabilityBufferSchema,
        ModelType.mrsort.name: ExplainabilityMRSortSchema,
        ModelType.weighted_sum.name: ExplainabilityWeightedSumSchema,
        ModelType.merge_overlap.name: ExplainabilityOverlapSchema,
        ModelType.dissolve_adjacent.name: ExplainabilityMergedFeaturesSchema,
        ModelType.zone_proposition.name: ExplainabilityZonePropositionSchema,
    }

    def get_obj_type(self, obj):
        return obj.model_type.name


class FeatureSchema(BaseSchema):
    class Meta(BaseSchema.Meta):
        model = Feature
        fields = (
            "id",
            "type",
            "data",
            "properties",
            "explainability",
            "execution_artifact",
            "input_data",
            "output_data",
        )
        # This schema is never use for creation nor update
        # so basically all fields are not updatable nor creatable (no write)
        dump_only = (
            "id",
            "type",
            "data",
            "properties",
            "explainability",
            "execution_artifact",
            "input_data",
            "output_data",
        )

    type = fields.Enum(FeatureType)
    properties = FieldMethod(
        fields.Dict(),
        "get_properties",
        metadata={
            "description": "feature properties organized by property name"
        },
    )
    explainability = FieldMethod(
        fields.Nested(ExplainabilitySchema), serialize="get_explainability"
    )
    execution_artifact = FieldMethod(
        fields.Nested(ExplainabilitySchema), serialize="get_execution_artifact"
    )
    input_data = FieldMethod(
        fields.Nested(BaseDataSchema, many=True),
        serialize="get_input_data",
    )
    output_data = FieldMethod(
        fields.Nested(BaseDataSchema, many=True),
        serialize="get_output_data",
    )
    data = FieldPluck(BaseData, "id")

    def get_input_data(self, obj):
        return BaseDataSchema(many=True).dump(obj.get_input_data())

    def get_output_data(self, obj):
        return BaseDataSchema(many=True).dump(obj.get_output_data())

    def get_properties(self, obj):
        return {p.attribute.name: p.value for p in obj.properties}

    def get_explainability(self, obj):
        return obj.explain()

    def get_execution_artifact(self, obj):
        return obj.execution_artifact


class BaseDataFeaturesSchema(BaseSchema):
    class Meta(BaseSchema.Meta):
        model = BaseData
        fields = ("features",)
        dump_only = ("features",)

    features = FieldPluck(Feature, "id", many=True)



from flask import send_file
from flask_restful import Resource, reqparse

from ..exceptions import RequestError
from ..models import ZoneProposition
from ..services.auth_service import token_required
from ..services.geo_data_loading_service import (
    get_geo_data_download,
    get_temporary_directory,
)
from .utils import check_permission, with_query_arg, with_response


class ZonePropositionAPI(Resource):
    @token_required
    @with_response(
        status=200,
        schema="#/components/schemas/FeatureCollection",
        description="Features of the recommended zone",
    )
    def get(self, proposition_id):
        """
        Return the details of the recommended zones.

        :param int proposition_id: The id for the recommended zones.
        :reqheader Authorization:
            JSON Web Token with Bearer scheme
            (``Authorization: Bearer <token>``).
        :status 403: The user is not allowed to access the project.
        :status 404: The resource doesn't exist.

        .. :quickref: Zone proposition; Get the details of a zone proposition
        """
        proposition = ZoneProposition.get_by_id(proposition_id)
        if proposition is None:
            raise RequestError("The zone proposition does not exist", 404)
        project = proposition.project
        check_permission(project)
        features = [f.as_geojson() for f in proposition.data.features]
        return {"features": features}


class ZonePropositionDownloadAPI(Resource):
    @token_required
    @with_query_arg(
        "outputFormat",
        str,
        required=False,
        description="Desired output format (leave blank for zip shapefile)",
    )
    @with_response(
        status=200,
        description="Zone Proposition as Geodata in specified file format",
        content={
            "application/octet-stream": {"type": "string", "format": "binary"}
        },
    )
    def get(self, proposition_id):
        """
        Return the recommended zones as geodata.

        :param int proposition_id: The id for the recommended zones.
        :reqheader Authorization:
            JSON Web Token with Bearer scheme
            (``Authorization: Bearer <token>``).
        :status 403: The user is not allowed to access the project.
        :status 404: The resource doesn't exist.
        :status 500:
            There was an error with the geometry while creating the file.

        .. :quickref: Zone proposition; Download zone proposition as geodata
        """
        parser = reqparse.RequestParser()
        parser.add_argument("outputFormat", type=str, location="args")
        args = parser.parse_args()
        proposition = ZoneProposition.get_by_id(proposition_id)
        if proposition is None:
            raise RequestError("The zone proposition does not exist", 404)
        project = proposition.project
        check_permission(project)
        with get_temporary_directory() as temp_dir:
            filename, file_object = get_geo_data_download(
                proposition.data, temp_dir, args.get("outputFormat")
            )
            return send_file(
                file_object, as_attachment=True, download_name=filename
            )



from ..exceptions import PermissionError, RequestError
from ..models import (
    DataAttribute,
    DataShare,
    GeneratedGeoData,
    Project,
    ProjectTaskModel,
    Template,
)
from ..models.project_data_utils import ProjectDataProcessingService
from .permission_service import has_permission


class ProjectService:
    __model__ = Project

    def get_if_authorized(self, project_id, user=None):
        """
        Return the project if the user has the permissions to view it.

        Args:
            project_id (int): Id of the project to return.
            user: User object.

        Raises:
            PermissionError if the user is not authorized to view this project.
        """
        project = self.__model__.get_by_id(project_id)
        if project is None:
            raise RequestError("The project does not exist", 404)
        if not has_permission(project, user):
            raise PermissionError
        return project

    def get_template_if_authorized(self, template_id, user=None):
        template = Template.get_by_id(template_id)
        if template is None:
            raise RequestError("The template does not exist", 404)
        if not has_permission(template, user):
            raise PermissionError
        return template

    def get_template_if_manager(self, template_id, user=None):
        assert user is not None
        template = self.get_template_if_authorized(template_id, user)
        if template.manager is not user or template.manager is None:
            raise PermissionError
        return template

    def get_all_authorized_projects(self, user=None):
        """
        Return the list of all project viewable by the user.

        The templates are not returned by this function.
        """
        project_list = Project.query.order_by(Project.id).all()
        return list(filter(lambda p: has_permission(p, user), project_list))

    def get_if_manager(self, project_id, user):
        """Return the project if the user if the project's manager."""
        assert user is not None
        project = self.get_if_authorized(project_id, user)
        if project.manager is not user or project.manager is None:
            raise PermissionError
        return project

    def get_user_templates(self, user=None):
        """Return the templates of the user."""
        template_list = Template.query.order_by(Template.id).all()
        return list(filter(lambda p: has_permission(p, user), template_list))

    def get_managed_templates(self, user):
        """Return the managed templates of the user."""
        query = Template.query.filter_by(manager=user)
        return query.order_by(Template.id).all()

    def get_project_tasks(self, project_id, limit=None, offset=None):
        """Return the list of tasks started in the project.

        Args:
            project_id (int): The id of the project.
            limit (int):
                The number of tasks to return. Default: None (all the tasks).
            offset (int):
                Offset for the tasks list.
                Default: None (start from the beginning).
        """
        project = self.get_if_authorized(project_id)
        query = ProjectTaskModel.query.filter_by(project_id=project.id)
        query = query.order_by(ProjectTaskModel.started_at.desc())
        if limit is not None:
            query = query.limit(limit)
        if offset is not None:
            query = query.offset(offset)
        return query.all()

    def get_total_project_tasks(self, project_id):
        """Return the total number of tasks in this project."""
        project = self.get_if_authorized(project_id)
        query = ProjectTaskModel.query.filter_by(project_id=project.id)
        return query.count()

    def create_template_from_project(self, project_id, user):
        """Create a template from a project."""
        project = self.get_if_authorized(project_id, user)
        template = Template.from_project(project, user)
        return template

    def create_project_from_template(self, template_id, manager):
        """
        Create a new projet from a JSON template.

        Args:
            template (std): The JSON template of the project. Will be modified.
            manager (User): The user that will be the manager of the project.
        """
        template = self.get_template_if_authorized(template_id, manager)
        project = Project.from_template(template, manager)
        return project

    def replace_input_attributes(
        self,
        project_id,
        attributes_mapping: dict[DataAttribute, DataAttribute],
    ) -> Project:
        """Replace input attributes of project data by new ones.

        :param project_id:
        :param attributes_mapping: mapping of old attributes to new ones
        :raise ValueError:
            * if project data graph cycles would be introduced
            * if some attributes are not in any data of the project
            * if some new attributes introduced were already in the data inputs
        :raise TypeError: if an attribute has changed its type
        :return: modified project
        """
        project = self.get_if_authorized(project_id)
        ps = ProjectDataProcessingService()

        old_attributes = set(attributes_mapping.keys())
        for data in project.data_list:
            to_replace_atts = (
                data._get_used_input_attributes() & old_attributes
            )
            if len(to_replace_atts) == 0:
                continue
            ps.replace_input_attributes(data, attributes_mapping)
        return project

    def get_authorized_shares(self, project_id: int) -> list[DataShare]:
        """Get all project data shares which the user has authorization for.

        :param project_id:
        :return: authorized data shares
        """
        project = self.get_if_authorized(project_id)
        shared = []
        for data in project.data_list:
            if data.data is None:
                continue
            match data.data:
                case GeneratedGeoData():
                    pass
                case _:
                    if not has_permission(data.data):
                        continue
            for share in data.data.shares:
                if not share.is_expired():
                    shared.append(share)
        return shared



from importlib import resources

from flask import Flask
from flask_migrate import Migrate

from ..models import db

migrate = Migrate()


def init_app(app: Flask):
    """Create the migrate group of commands."""
    with resources.path("deseasion.backend", "migrations") as migrations_path:
        migrate.init_app(app, db, str(migrations_path))



import enum
import io
import tokenize
from collections import defaultdict
from copy import deepcopy

import numpy as np
from sqlalchemy.dialects.postgresql import ARRAY, JSONB
from sqlalchemy.ext.orderinglist import ordering_list
from sqlalchemy.orm import backref, validates

from ..exceptions import GeometryTypeError, ProcessingError
from . import db
from .geo_data import (
    AttributeType,
    DataAttribute,
    DataAttributeOrdinal,
    DataValueNominal,
    DataValueQuantitative,
    Feature,
)
from .mixins import BaseModelMixin, ModelMixin
from .processing_model_utils import evaluate_discrete_rule


class KeepOverlap(enum.Enum):
    min = 1
    max = 2
    sum = 3
    average = 4


class ModelType(enum.Enum):
    categories_rule = 1
    continuous_rule = 2
    geo_buffer = 3
    mrsort = 4
    weighted_sum = 5
    merge_overlap = 6
    dissolve_adjacent = 7
    zone_proposition = 8


class ProcessingModel(ModelMixin, db.Model):
    __tablename__ = "processing_model"

    model_type = db.Column(db.Enum(ModelType))
    name = db.Column(db.String)
    data_generator_id = db.Column(
        db.Integer, db.ForeignKey("project_data.id"), nullable=False
    )
    cut_to_extent = db.Column(db.Boolean)

    data_generator = db.relationship(
        "DataGenerator",
        back_populates="_processing_models",
        foreign_keys=[data_generator_id],
    )

    __mapper_args__ = {"polymorphic_on": model_type}

    def __init__(
        self,
        cut_to_extent=True,
        name="",
        **kwargs,
    ):
        self.cut_to_extent = cut_to_extent
        self.name = name

    def __deepcopy__(self, memo):
        other = type(self)()
        memo[id(self)] = other
        other.name = self.name
        other.cut_to_extent = self.cut_to_extent
        other.data_generator = deepcopy(self.data_generator, memo)
        return other

    @property
    def input_data(self):
        return self.data_generator.input_data

    def _get_used_input_attributes(self) -> set[DataAttribute]:
        """Get set of used input attributes in a model definition.

        Here it retuns all input attributes.

        :return:
        """
        return {
            attr
            for data in self.input_data
            if data.data is not None
            for attr in data.data.attributes
        }

    def get_used_input_attributes(self) -> list[DataAttribute]:
        """Get list of used input attributes in a model definition.

        :return:
        """
        return sorted(
            self._get_used_input_attributes(), key=lambda a: (a.data.id, a.id)
        )

    def explain(self, feature: Feature) -> dict:
        """Explain feature wrt the model.

        :param feature: feature computed by the model
        :return: explanation as a dict
        """
        explanation = {
            "model_type": self.model_type.name,
            "disaggregation": [],
        }
        for input_feature in feature.input_features:
            for prop in input_feature.properties:
                if prop.attribute not in self._get_used_input_attributes():
                    continue
                if prop.value is not None:
                    explanation["disaggregation"].append(
                        {
                            "data": prop.attribute.data.name,
                            "attribute": prop.attribute.name,
                            "value": prop.value,
                        }
                    )
        return explanation


class MergeOverlapModel(ProcessingModel):
    keep_overlap = db.Column(
        db.Enum(KeepOverlap), default=KeepOverlap.max, nullable=False
    )

    __mapper_args__ = {"polymorphic_identity": ModelType.merge_overlap}

    def __init__(self, keep_overlap=KeepOverlap.max, **kwargs):
        super().__init__(**kwargs)
        self.keep_overlap = keep_overlap

    def get_overlap_value(self, values):
        """Return the overlap value filtered through the overlap function.

        Args:
            values: A list of the values.
        """
        if isinstance(values[0], DataValueNominal):
            raise TypeError("Cannot aggregate overlapping nominal values")
        if self.keep_overlap == KeepOverlap.max:
            return max(values).value
        if self.keep_overlap == KeepOverlap.min:
            return min(values).value
        if not isinstance(values[0], DataValueQuantitative):
            raise TypeError(
                "Ordinal values can only be aggregated with "
                f"keep_overlap in ['min', 'max'], found {self.keep_overlap}"
            )
        if self.keep_overlap == KeepOverlap.sum:
            return sum(values).value
        if self.keep_overlap == KeepOverlap.average:
            return sum(values) / len(values).value

    def process_overlap(self, entity):
        """Decompose the overlapping attributes, and keep the chosen value
        (all, worse or best)"""
        try:
            result = entity.decompose_intersection()
        except GeometryTypeError:
            raise ProcessingError(
                "The overlapping geometries are not polygons."
            )
        for feature in result.features:
            for prop, values in feature.data.attributes.items():
                values = [v for v in values if v is not None]
                if values is None or len(values) < 1:
                    val = values
                elif len(values) == 1:
                    val = values[0]
                else:
                    try:
                        val = self.get_overlap_value(values)
                    except TypeError:
                        raise ProcessingError(
                            "Error overlapping the value types."
                        )
                feature.data.attributes[prop] = val
        return result

    def explain(self, feature):
        """Explain feature wrt the model.

        :param feature: feature computed by the model
        :return: explanation as a dict
        """
        explanation = {
            "model_type": self.model_type.name,
            "keep_overlap": self.keep_overlap.name,
            "aggregated": [],
        }
        tmp_values = defaultdict(list)
        for input_feature in feature.input_features:
            for prop in input_feature.properties:
                if prop.attribute not in self._get_used_input_attributes():
                    continue
                if prop.value is not None:
                    tmp_values[prop.attribute].append(prop.value)
        for attr, vals in tmp_values.items():
            if len(vals) > 0:
                explanation["aggregated"].append(
                    {
                        "data": attr.data.name,
                        "attribute": attr.name,
                        "values": vals,
                    }
                )
        return explanation

    def __deepcopy__(self, memo):
        other = super().__deepcopy__(memo)
        other.keep_overlap = self.keep_overlap
        return other


class DissolveAdjacentModel(ProcessingModel):
    __mapper_args__ = {"polymorphic_identity": ModelType.dissolve_adjacent}

    def explain(self, feature):
        """Explain feature wrt the model.

        :param feature: feature computed by the model
        :return: explanation as a dict
        """
        return {"model_type": self.model_type.name}


def _extract_attributes_from_code(
    code: str, model: ProcessingModel
) -> set[DataAttribute]:
    """Extract attributes from code.

    :param code:
    :param model: processing model
    :return: set of attributes used in code
    """
    imput_attributes = {}
    for data in model.input_data:
        if data.data is None:
            continue
        for attribute in data.data.attributes:
            imput_attributes[(data.name, attribute.name)] = attribute
    tokens = list(tokenize.generate_tokens(io.StringIO(code).readline))
    res = set()
    i = 0
    while i < len(tokens) - 2:
        token = tokens[i]
        if (
            token.type == tokenize.NAME
            and tokens[i + 1].type == tokenize.OP
            and tokens[i + 1].string == "."
            and tokens[i + 2].type == tokenize.NAME
            and (token.string, tokens[i + 2].string) in imput_attributes
        ):
            res.add(imput_attributes[(token.string, tokens[i + 2].string)])
            i += 3
        else:
            i += 1

    return res


def _replace_variables_code(
    code: str,
    variables_mapping: dict[tuple[str, str], tuple[str, str]],
) -> str:
    """Replace attribute variables in code by new ones.

    :param code:
    :param variables_mapping: mapping of old variables to new ones
    :return: new code
    """
    tokens = list(tokenize.generate_tokens(io.StringIO(code).readline))
    res = []
    i = 0
    while i < len(tokens):
        token = tokens[i]
        if (
            token.type == tokenize.NAME
            and i < len(tokens) - 2
            and tokens[i + 1].type == tokenize.OP
            and tokens[i + 1].string == "."
            and tokens[i + 2].type == tokenize.NAME
        ):
            new_data, new_attr = variables_mapping.get(
                (token.string, tokens[i + 2].string),
                (token.string, tokens[i + 2].string),
            )
            res.append(
                tokenize.TokenInfo(
                    type=tokenize.NAME,
                    string=new_data,
                    start=token.start,
                    end=token.end,
                    line=token.line,
                )
            )
            res.append(tokens[i + 1])
            res.append(
                tokenize.TokenInfo(
                    type=tokenize.NAME,
                    string=new_attr,
                    start=tokens[i + 2].start,
                    end=tokens[i + 2].end,
                    line=tokens[i + 2].line,
                )
            )
            i += 3
        else:
            res.append(token)
            i += 1

    return tokenize.untokenize(res)


class DefaultValue(BaseModelMixin, db.Model):
    attribute_id = db.Column(
        db.Integer,
        db.ForeignKey("data_attribute.id", ondelete="CASCADE"),
        nullable=False,
        primary_key=True,
    )
    model_id = db.Column(
        db.Integer,
        db.ForeignKey("processing_model.id", ondelete="CASCADE"),
        nullable=False,
        primary_key=True,
    )
    value = db.Column(JSONB)

    attribute = db.relationship("DataAttribute")
    model = db.relationship(
        "PrefDefaultValues", back_populates="default_values"
    )

    def __deepcopy__(self, memo):
        other = type(self)()
        memo[id(self)] = other
        other.attribute = deepcopy(self.attribute, memo)
        other.model = deepcopy(self.model, memo)
        other.value = self.value
        return other


class PrefDefaultValues(ProcessingModel):
    default_values = db.relationship(
        "DefaultValue", back_populates="model", cascade="all,delete-orphan"
    )

    def __deepcopy__(self, memo):
        other = super().__deepcopy__(memo)
        other.default_values = [deepcopy(d, memo) for d in self.default_values]
        return other

    # @validates('default_values')
    # def validate_default_values(self, key, default_value):
    # TODO: Check that the attribute is in the input data of the model
    # pass

    def get_default_value(self, attribute):
        """Returns the default value for the attribute.

        Returns:
            The default value if it exists, or None otherwise
        """
        try:
            return next(
                d.value
                for d in self.default_values
                if d.attribute_id == attribute.id
            )
        except StopIteration:
            return None

    def explain(self, feature):
        """Explain feature wrt the model.

        :param feature: feature computed by the model
        :return: explanation as a dict
        """
        explanation = {
            "model_type": self.model_type.name,
            "disaggregation": [],
        }
        input_map = {
            f.data: {prop.attribute: prop.value for prop in f.properties}
            for f in feature.input_features
        }
        for attr in self.get_used_input_attributes():
            value = self.get_default_value(attr)
            if (
                attr.data in input_map
                and attr in input_map[attr.data]
                and input_map[attr.data][attr] is not None
            ):
                value = input_map[attr.data][attr]
            if value is not None:
                explanation["disaggregation"].append(
                    {
                        "data": attr.data.name,
                        "attribute": attr.name,
                        "value": value,
                    }
                )
        return explanation


class GeoBuffer(ProcessingModel):
    """Model to create a buffer around the geometry."""

    radius = db.Column(db.BigInteger)

    __mapper_args__ = {"polymorphic_identity": ModelType.geo_buffer}

    def __init__(self, radius=None, **kwargs):
        super().__init__(**kwargs)
        self.radius = radius

    def __deepcopy__(self, memo):
        other = super().__deepcopy__(memo)
        other.radius = self.radius
        return other

    def explain(self, feature):
        """Explain feature wrt the model.

        :param feature: feature computed by the model
        :return: explanation as a dict
        """
        explanation = super().explain(feature)
        explanation["radius"] = self.radius
        return explanation


class DiscreteCategory(ModelMixin, db.Model):
    """Model for creating data using discrete values (categories)."""

    __tablename__ = "discrete_category"

    type = db.Column(db.String)
    name = db.Column(db.String)
    position = db.Column(db.Integer)
    preference_model_id = db.Column(
        db.Integer, db.ForeignKey("processing_model.id", ondelete="CASCADE")
    )

    preference_model = db.relationship("DiscreteModel")

    __mapper_args__ = {
        "polymorphic_on": type,
        "polymorphic_identity": "discrete_category",
    }

    def __init__(self, name, **kwargs):
        super().__init__(**kwargs)
        self.name = name

    def __deepcopy__(self, memo):
        other = type(self)(name=self.name, position=self.position)
        memo[id(self)] = other
        other.preference_model = deepcopy(self.preference_model, memo)
        return other


class DiscreteRulesCategory(DiscreteCategory):
    rules = db.Column(ARRAY(db.String, dimensions=1))

    __mapper_args__ = {"polymorphic_identity": "discrete_rules_category"}

    def __init__(self, rules=[], **kwargs):
        super().__init__(**kwargs)
        self.rules = rules

    def __deepcopy__(self, memo):
        other = type(self)(name=self.name)
        memo[id(self)] = other
        other.preference_model = deepcopy(self.preference_model, memo)
        other.rules = deepcopy(self.rules, memo)
        return other

    def evaluate_data(self, data, stats=None):
        evaluation = np.full_like(
            data,
            fill_value=False,
            dtype=bool,
            shape=(len(self.rules), len(data)),
        )
        for i, rule in enumerate(self.rules):
            evaluation[i, :] = evaluate_discrete_rule(data, rule, stats)
        return evaluation


class DiscreteModel(ProcessingModel):
    categories = db.relationship(
        "DiscreteCategory",
        order_by="DiscreteCategory.position",
        cascade="all,delete-orphan",
        collection_class=ordering_list("position"),
        back_populates="preference_model",
    )


class DiscreteRules(DiscreteModel):
    """Used to create a geo-data containing discrete values (ie. categories)
    from rules."""

    categories = db.relationship(
        "DiscreteRulesCategory",
        order_by="DiscreteRulesCategory.position",
        cascade="all,delete-orphan",
        collection_class=ordering_list("position"),
        back_populates="preference_model",
    )

    __mapper_args__ = {"polymorphic_identity": ModelType.categories_rule}

    def __init__(self, categories=[], **kwargs):
        super().__init__(**kwargs)
        self.categories = categories

    def __deepcopy__(self, memo):
        other = super().__deepcopy__(memo)
        other.categories = [deepcopy(c, memo) for c in self.categories]
        return other

    def _get_used_input_attributes(self) -> set[DataAttribute]:
        """Get set of used input attributes in a model definition.

        :return:
        """
        attributes = set()
        for category in self.categories:
            for rule in category.rules:
                attributes |= _extract_attributes_from_code(rule, self)
        return attributes

    def explain(self, feature):
        """Explain feature wrt the model.

        :param feature: feature computed by the model
        :return: explanation as a dict
        """
        explanation = super().explain(feature)
        explanation["rule"] = feature.execution_artifact["rule"]
        return explanation


class MRSort(DiscreteModel):
    """Used to create a geo-data using the MR-Sort algorithm.

    Attributes:
        categories (list): The categories for the sorting algorithm.
            Best category first.
        majority_threshold (float): The cut threshold for the concordance
            condition. At least half the sum of the criteria weights.
    """

    majority_threshold = db.Column(db.Float)

    criteria = db.relationship(
        "MRSortCriterion", back_populates="mrsort", cascade="all,delete-orphan"
    )

    __mapper_args__ = {"polymorphic_identity": ModelType.mrsort}

    def __init__(self, criteria=[], categories=[], **kwargs):
        super().__init__(**kwargs)
        self.criteria = criteria
        self.categories = categories

    def __deepcopy__(self, memo):
        other = super().__deepcopy__(memo)
        other.majority_threshold = self.majority_threshold
        other.criteria = [deepcopy(c, memo) for c in self.criteria]
        other.categories = [deepcopy(c, memo) for c in self.categories]
        return other

    @validates("criteria")
    def validate_criteria(self, key, criterion):
        """Verify that the criterion references an input_data."""
        attributes = []
        for data in self.input_data:
            if data.data is not None:
                attributes += data.data.attributes
        assert (
            criterion.attribute in attributes
        ), "The criterion references an invalid data."
        return criterion

    def init_criteria(self):
        criteria = []
        for data in self.input_data:
            if data.data is not None:
                for attr in data.data.attributes:
                    if attr.type is AttributeType.quantitative:
                        for crit in self.criteria:
                            if crit.attribute is attr:
                                criterion = crit
                                break
                        else:
                            criterion = MRSortCriterion(
                                attribute=attr, mrsort_id=self.id
                            )
                        criteria.append(criterion)
        self.criteria = criteria

    def compute_weights(self, attributes, profile_index, criteria_lookup=None):
        """Compute weights where feature is better than profile.

        :param attributes: list of (data name, attribute name, value)
        :param profile_index: index of profile to compare to
        :param criteria_lookup:
            dict mapping (data name, attribute name) to criteria,
            computed if absent
        :return:
            weights for attributes where feature is better than profile
            indexed by (data name, attribute name)
        """
        criteria_lookup = criteria_lookup or {
            (c.attribute.data.name, c.attribute.name): c for c in self.criteria
        }
        weights = {}
        for name, attribute, value in attributes:
            criterion = criteria_lookup.get((name, attribute), None)
            if criterion is not None and criterion.is_better(
                value, profile_index
            ):
                weights[(name, attribute)] = criterion.weight
        return weights

    def check_category(self, attributes):
        """Execute the MR-Sort algorithm on the attributes.

        Args:
            attributes (list): List of ('name', 'attribute', value).

        Raises:
            AttributeError: If the attribute ('name', 'attribute') does not
                correspond to any criterion.
        """
        categories = self.categories[::-1]  # reverse the order (worst first)
        result = categories[0]
        criteria_lookup = {
            (c.attribute.data.name, c.attribute.name): c for c in self.criteria
        }
        for index, category in enumerate(categories[1:]):
            weights = self.compute_weights(attributes, index, criteria_lookup)
            total = sum(weights.values())
            if total >= self.majority_threshold:
                result = category
            else:
                break
        return result

    def _get_used_input_attributes(self) -> set[DataAttribute]:
        """Get set of used input attributes in a model definition.

        :return:
        """
        return {
            criterion.attribute
            for criterion in self.criteria
            if criterion.attribute is not None
        }

    def explain(self, feature):
        """Explain feature wrt the model.

        :param feature: feature computed by the model
        :return: explanation as a dict
        """
        explanation = super().explain(feature)
        explanation["majority_threshold"] = self.majority_threshold
        category_name = feature.get_property("category")
        attributes = []
        for value in explanation["disaggregation"]:
            attributes.append(
                (value["data"], value["attribute"], value["value"])
            )
        lower = None
        upper = None
        explanation["categories"] = []
        for index, cat in enumerate(
            sorted(self.categories, key=lambda c: c.position)
        ):
            explanation["categories"].append(cat.name)
            if cat.name != category_name:
                continue
            if index > 0:
                upper = len(self.categories) - index - 1
            if index < len(self.categories) - 1:
                lower = len(self.categories) - index - 2
        if lower is not None:
            explanation["lower"] = []
            lower_weights = self.compute_weights(attributes, lower)
        if upper is not None:
            explanation["upper"] = []
            upper_weights = self.compute_weights(attributes, upper)

        explanation["criteria"] = []
        for crit in self.criteria:
            explanation["criteria"].append(
                {
                    "data": crit.attribute.data.name,
                    "attribute": crit.attribute.name,
                    "maximize": crit.maximize,
                }
            )
            if lower is not None:
                explanation["lower"].append(
                    {
                        "data": crit.attribute.data.name,
                        "attribute": crit.attribute.name,
                        "value": crit.profiles[lower],
                        "weight": lower_weights.get(
                            (crit.attribute.data.name, crit.attribute.name), 0
                        ),
                    }
                )
            if upper is not None:
                explanation["upper"].append(
                    {
                        "data": crit.attribute.data.name,
                        "attribute": crit.attribute.name,
                        "value": crit.profiles[upper],
                        "weight": upper_weights.get(
                            (crit.attribute.data.name, crit.attribute.name), 0
                        ),
                    }
                )

        return explanation


class MRSortCriterion(BaseModelMixin, db.Model):
    """Contain the criterion data for the MR-Sort algorithm.

    Attributes:
        profiles: The values for the profiles of the categories.
            Worst value first.
        weight (float):
            The weight of the criterion in the algorithm.
        maximize (bool):
            Preference direction (maximize if True, minimize if False).
            Default: True.
    """

    __tablename__ = "mrsort_criterion"

    profiles = db.Column(ARRAY(db.Float, dimensions=1))
    weight = db.Column(db.Float)
    maximize = db.Column(db.Boolean, nullable=False, default=False)
    mrsort_id = db.Column(
        db.Integer,
        db.ForeignKey("processing_model.id", ondelete="CASCADE"),
        primary_key=True,
    )
    attribute_id = db.Column(
        db.Integer,
        db.ForeignKey("data_attribute.id", ondelete="CASCADE"),
        primary_key=True,
    )

    attribute = db.relationship(
        "DataAttribute",
        backref=backref("mrsort_criteria", cascade="delete,delete-orphan"),
    )
    mrsort = db.relationship("MRSort", back_populates="criteria")

    def __init__(
        self, mrsort=None, profiles=[], weight=0, maximize=True, **kwargs
    ):
        super().__init__(**kwargs)
        self.profiles = profiles
        self.weight = weight
        self.maximize = maximize

    def __deepcopy__(self, memo):
        other = type(self)()
        memo[id(self)] = other
        other.profiles = deepcopy(self.profiles, memo)
        other.weight = self.weight
        other.maximize = self.maximize
        other.attribute = deepcopy(self.attribute, memo)
        other.mrsort = deepcopy(self.mrsort, memo)
        return other

    def is_better(self, value, profile_index):
        """Check if the value is better than the profile for the given index.

        Args:
            value (number): The value to check.
            profile_index (int): Index of the profile in the profiles list.
        """
        if value is None:
            return False
        if self.maximize:
            return value >= self.profiles[profile_index]
        else:
            return value <= self.profiles[profile_index]


class ContinuousRule(PrefDefaultValues):
    """Used to create a geo-data containing continuous values from a rule"""

    rule = db.Column(db.String)

    __mapper_args__ = {"polymorphic_identity": ModelType.continuous_rule}

    def __init__(self, rule="", default_values=[], **kwargs):
        super().__init__(**kwargs)
        self.rule = rule
        self.default_values = default_values
        self.model_type = ModelType.continuous_rule

    def __deepcopy__(self, memo):
        other = super().__deepcopy__(memo)
        other.rule = self.rule
        return other

    def _get_used_input_attributes(self) -> set[DataAttribute]:
        """Get set of used input attributes in a model definition.

        :return:
        """
        return _extract_attributes_from_code(self.rule, self)


class WeightedSum(PrefDefaultValues):
    operands = db.relationship(
        "WeightedSumOperand",
        back_populates="model",
        cascade="all,delete-orphan",
    )

    __mapper_args__ = {"polymorphic_identity": ModelType.weighted_sum}

    def __init__(self, operands=None, **kwargs):
        super().__init__(**kwargs)
        if operands is None:
            self.operands = [WeightedSumOperand(weight=1)]
        else:
            self.operands = operands

    def __deepcopy__(self, memo):
        other = super().__deepcopy__(memo)
        other.operands = [deepcopy(o, memo) for o in self.operands]
        return other

    def _get_used_input_attributes(self) -> set[DataAttribute]:
        """Get set of used input attributes in a model definition.

        :return:
        """
        return {
            op.attribute for op in self.operands if op.attribute is not None
        }

    def explain(self, feature):
        """Explain feature wrt the model.

        :param feature: feature computed by the model
        :return: explanation as a dict
        """
        explanation = super().explain(feature)
        properties = {
            (p["data"], p["attribute"]): p["value"]
            for p in explanation.pop("disaggregation")
        }
        default_values = {
            (d.attribute.data.name, d.attribute.name): d.value
            for d in self.default_values
        }
        explanation["disaggregation"] = []
        for operand in self.operands:
            index = (operand.attribute.data.name, operand.attribute.name)
            val = properties.get(index)
            if val is None:
                val = default_values.get(index)
            if val is not None:
                explanation["disaggregation"].append(
                    {
                        "data": operand.attribute.data.name,
                        "attribute": operand.attribute.name,
                        "value": val,
                        "weight": operand.weight,
                    }
                )
        return explanation


class WeightedSumOperand(ModelMixin, db.Model):
    __tablename__ = "weighted_sum_operand"

    attribute_id = db.Column(
        db.Integer, db.ForeignKey("data_attribute.id", ondelete="SET NULL")
    )
    weight = db.Column(db.Float)
    model_id = db.Column(
        db.Integer,
        db.ForeignKey("processing_model.id", ondelete="CASCADE"),
        nullable=False,
    )

    attribute = db.relationship("DataAttribute")
    model = db.relationship("WeightedSum", back_populates="operands")

    def __deepcopy__(self, memo):
        other = type(self)()
        memo[id(self)] = other
        other.attribute = deepcopy(self.attribute, memo)
        other.model = deepcopy(self.model, memo)
        other.weight = self.weight
        return other


class ZonePropositionGenerator(ProcessingModel):
    geo_size = db.Column(db.Float)
    iterations = db.Column(db.Integer)
    duration = db.Column(db.Float)

    size = db.Column(db.Integer)
    mutation = db.Column(db.Float)
    children = db.Column(db.Integer)
    filter_clusters = db.Column(db.Boolean)

    __mapper_args__ = {"polymorphic_identity": ModelType.zone_proposition}

    def __init__(
        self,
        geo_size=0,
        size=80,
        mutation=0.02,
        children=60,
        filter_clusters=True,
        iterations=None,
        duration=None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.geo_size = geo_size
        self.size = size
        self.mutation = mutation
        self.children = children
        self.filter_clusters = filter_clusters
        self.iterations = iterations
        self.duration = duration

    def __deepcopy__(self, memo):
        other = type(self)()
        memo[id(self)] = other
        other.name = self.name
        other.cut_to_extent = self.cut_to_extent
        other.geo_size = self.geo_size
        other.iterations = self.iterations
        other.duration = self.duration
        other.size = self.size
        other.mutation = self.mutation
        other.children = self.children
        other.filter_clusters = self.filter_clusters
        other.data_generator = deepcopy(self.data_generator, memo)
        return other

    def ga_params(self):
        return {
            "size": self.size,
            "mutation": self.mutation,
            "children": self.children,
            "filter_clusters": self.filter_clusters,
        }

    def _get_used_input_attributes(self) -> set[DataAttribute]:
        """Get set of used input attributes in a model definition.

        Here it retuns all input attributes.

        :return:
        """
        if len(self.input_data) == 0:
            return set()
        input_data = self.input_data[0]
        if input_data.data is None:
            return set()
        if len(input_data.data.attributes) == 0:
            return set()
        return {input_data.data.attributes[0]}

    def explain(self, feature):
        """Explain feature wrt the model.

        :param feature: feature computed by the model
        :return: explanation as a dict
        """
        explanation = {
            "model_type": self.model_type.name,
            "disaggregation": [],
        }
        _attributes = self.get_used_input_attributes()
        attribute = None if len(_attributes) == 0 else _attributes[0]
        is_ordinal_attribute = isinstance(attribute, DataAttributeOrdinal)
        feature_artifacts = {
            artifact["feature"]: artifact
            for artifact in feature.execution_artifact["features"]
        }
        for input_feature in feature.input_features:
            area = feature_artifacts[input_feature.id]["area"]
            for prop in input_feature.properties:
                if prop.attribute != attribute:
                    continue
                if prop.value is not None:
                    value = float(
                        len(attribute.order)
                        - attribute.order.index(prop.value)
                        if is_ordinal_attribute
                        else prop.value
                    )
                    explanation["disaggregation"].append(
                        {
                            "data": attribute.data.name,
                            "attribute": attribute.name,
                            "value": prop.value,
                            "numeric_value": value,
                            "area": area,
                            "fitness": feature_artifacts[input_feature.id][
                                "fitness"
                            ],
                        }
                    )
        return explanation



from flask_sqlalchemy import SQLAlchemy
from sqlalchemy import MetaData

convention = {
    "ix": "ix_%(column_0_label)s",
    "uq": "uq_%(table_name)s_%(column_0_name)s",
    "ck": "ck_%(table_name)s_%(constraint_name)s",
    "fk": "fk_%(table_name)s_%(column_0_name)s_%(referred_table_name)s",
    "pk": "pk_%(table_name)s",
}

metadata = MetaData(naming_convention=convention)
db = SQLAlchemy(None, metadata=metadata)

from .geo_data import (  # noqa: E402
    AttributeType,
    BaseData,
    BaseGeoData,
    DataAttribute,
    DataAttributeNominal,
    DataAttributeOrdinal,
    DataAttributeQuantitative,
    DataValue,
    Feature,
    FeatureType,
    GeneratedGeoData,
    GeoData,
    GeoDataType,
    GeoFeature,
    GlobalData,
    StreamGeoData,
    WFSGeoData,
    WMSGeoData,
)
from .permission import PermissionAbility, UserPermission  # noqa: E402
from .processing_models import (  # noqa: E402
    ContinuousRule,
    DiscreteCategory,
    DiscreteRules,
    DiscreteRulesCategory,
    GeoBuffer,
    KeepOverlap,
    ModelType,
    MRSort,
    MRSortCriterion,
    ProcessingModel,
)
from .project import Project, Template  # noqa: E402
from .project_data import (  # noqa: E402
    DataGenerator,
    DataGeo,
    DataStream,
    DataType,
    DynamicData,
    ProjectData,
    ProjectGlobalData,
)
from .share import DataShare  # noqa: E402
from .tasks import ProjectTaskModel, ProjectTaskType  # noqa: E402
from .user import User  # noqa: E402

__all__ = [
    "AttributeType",
    "BaseData",
    "BaseGeoData",
    "DataAttribute",
    "DataAttributeNominal",
    "DataAttributeOrdinal",
    "DataAttributeQuantitative",
    "DataValue",
    "Feature",
    "FeatureType",
    "GeneratedGeoData",
    "GeoData",
    "GeoDataType",
    "GeoFeature",
    "GlobalData",
    "GeoDataType",
    "StreamGeoData",
    "WFSGeoData",
    "WMSGeoData",
    "PermissionAbility",
    "UserPermission",
    "ContinuousRule",
    "DiscreteCategory",
    "DiscreteRules",
    "DiscreteRulesCategory",
    "GeoBuffer",
    "KeepOverlap",
    "MRSort",
    "MRSortCriterion",
    "ProcessingModel",
    "ModelType",
    "Project",
    "Template",
    "DataGenerator",
    "DataGeo",
    "DataType",
    "ProjectData",
    "ProjectGlobalData",
    "DataStream",
    "DynamicData",
    "DataShare",
    "ProjectTaskModel",
    "ProjectTaskType",
    "User",
]



import importlib.resources as pkg_resources

import click
from alembic import command
from alembic.migration import MigrationContext
from flask import Flask
from sqlalchemy.sql import text

from .. import sql
from ..models import db


def register_commands(app: Flask):
    db_cli_group = app.cli.commands["db"]

    @db_cli_group.command()
    @click.option(
        "-d",
        "--directory",
        default=None,
        help=('Migration script directory (default is "migrations")'),
    )
    @click.option(
        "--tag",
        default=None,
        help=(
            'Arbitrary "tag" name - can be used by custom env.py ' "scripts"
        ),
    )
    @click.option(
        "-f",
        "--force",
        default=False,
        is_flag=True,
        help="Force database initialization (even if already initialized)",
    )
    @click.option(
        "-x",
        "--x-arg",
        multiple=True,
        help="Additional arguments consumed by custom env.py scripts",
    )
    def create(directory, tag, force, x_arg):
        """Creates all the SQLAlchemy tables and add SQL functions."""

        config = app.extensions["migrate"].migrate.get_config(
            directory, x_arg=x_arg
        )

        engine = db.get_engine()
        with engine.connect() as conn:
            migration_ctx = MigrationContext.configure(conn)
            current = migration_ctx.get_current_revision()

        if current is None or force:
            db.create_all()
            with (
                pkg_resources.files(sql)
                .joinpath("dissolve_adjacent.sql")
                .open("r") as f
            ):
                db.session.execute(text(f.read()))
            with (
                pkg_resources.files(sql)
                .joinpath("tile_bbox.sql")
                .open("r") as f
            ):
                db.session.execute(text(f.read()))
            with (
                pkg_resources.files(sql)
                .joinpath("trigger_feature_make_valid.sql")
                .open("r") as f
            ):
                db.session.execute(text(f.read()))
            db.session.commit()
            command.stamp(config, revision="head", tag=tag)
        else:
            print(
                "Database already exists. You may force its creation "
                "with '--force' option (may break database).\n"
                "If wanting to upgrade it, use 'flask db upgrade' command "
                "instead."
            )

    @db_cli_group.command()
    @click.option(
        "-d",
        "--directory",
        default=None,
        help=('Migration script directory (default is "migrations")'),
    )
    @click.option(
        "-x",
        "--x-arg",
        multiple=True,
        help="Additional arguments consumed by custom env.py scripts",
    )
    def drop(directory, x_arg):
        """Drops all the SQLAlchemy tables."""

        config = app.extensions["migrate"].migrate.get_config(
            directory, x_arg=x_arg
        )

        if input("Your data will be lost. Continue? (Y/y): ")[0] in ["Y", "y"]:
            db.drop_all()
            command.stamp(config, revision=None)



"""Add data lineage to features

Revision ID: c6bfa5794b5c
Revises: 32025d48c84e
Create Date: 2025-02-19 17:40:03.972024

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "c6bfa5794b5c"
down_revision = "32025d48c84e"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "feature_input_association",
        sa.Column("feature_id", sa.Integer(), nullable=False),
        sa.Column("input_feature_id", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(
            ["feature_id"],
            ["feature.id"],
            name=op.f("fk_feature_input_association_feature_id_feature"),
        ),
        sa.ForeignKeyConstraint(
            ["input_feature_id"],
            ["feature.id"],
            name=op.f("fk_feature_input_association_input_feature_id_feature"),
        ),
        sa.PrimaryKeyConstraint(
            "feature_id",
            "input_feature_id",
            name=op.f("pk_feature_input_association"),
        ),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("feature_input_association")
    # ### end Alembic commands ###



from marshmallow import fields, validates

from ..models import (
    BaseGeoData,
    DataGenerator,
    DataGeo,
    DataStream,
    DataType,
    ProjectData,
    ProjectGlobalData,
)
from .base import (
    BaseSchema,
    FieldMethod,
    FieldPluck,
    OneOfSchema,
    OneOfSchemaWithType,
    constant_enum,
)
from .geo_data import BaseDataAttributeSchema, StreamGeoDataSchema
from .processing_models import (
    PrefModelBaseSchema,
    ProcessingModelCreationSchema,
)
from .project import ProjectSchema
from .utils import validate_name


class ProjectDataBaseSchema(BaseSchema):
    class Meta(BaseSchema.Meta):
        model = ProjectData
        dump_only = (
            "id",
            "data_type",
            "data",
            "last_update",
            "project",
            "is_outdated",
            "used_input_attributes",
            "used_output_attributes",
        )

    project = fields.Nested(ProjectSchema)
    data = FieldPluck(BaseGeoData, "id")
    data_type = fields.Enum(DataType, required=True)
    attributes = FieldMethod(
        fields.Nested(
            BaseDataAttributeSchema,
            only=("id", "name", "type", "statistics"),
            many=True,
        ),
        "get_attributes_list",
    )
    input_data = FieldPluck(ProjectData, "id", many=True)
    output_data = FieldPluck(ProjectData, "id", many=True)
    used_input_attributes = FieldMethod(
        fields.Nested(
            BaseDataAttributeSchema,
            only=("id", "name", "data", "type"),
            many=True,
        ),
        "get_used_input_attributes",
    )
    used_output_attributes = FieldMethod(
        fields.Nested(
            BaseDataAttributeSchema,
            only=("id", "name", "data", "type"),
            many=True,
        ),
        "get_used_output_attributes",
    )
    is_outdated = FieldMethod(fields.Boolean(), "check_outdated")

    @validates("name")
    def validate_name(self, value, **kwargs):
        validate_name(value)

    def get_attributes_list(self, obj):
        return obj.get_attributes_list()

    def get_used_input_attributes(self, obj):
        return [
            {
                "id": attribute.id,
                "name": attribute.name,
                "data": {"id": attribute.data_id, "name": attribute.data.name},
                "type": attribute.type.name,
            }
            for attribute in obj.get_used_input_attributes()
        ]

    def get_used_output_attributes(self, obj):
        return [
            {
                "id": attribute.id,
                "name": attribute.name,
                "data": {"id": attribute.data_id, "name": attribute.data.name},
                "type": attribute.type.name,
            }
            for attribute in obj.get_used_attributes()
        ]

    def check_outdated(self, obj):
        return obj.is_outdated()


class ProjectDataBaseCreationSchema(ProjectDataBaseSchema):
    pass


class DataGeoSchema(ProjectDataBaseSchema):
    class Meta(ProjectDataBaseSchema.Meta):
        model = DataGeo
        fields = (
            "id",
            "data_type",
            "created_at",
            "modified_at",
            "last_update",
            "project",
            "description",
            "name",
            "data",
            "attributes",
            "input_data",
            "output_data",
            "used_input_attributes",
            "used_output_attributes",
            "data_id",
            "is_outdated",
        )
        dump_only = (
            "id",
            "data_type",
            "data",
            "used_input_attributes",
            "used_output_attributes",
            "last_update",
            "created_at",
            "modified_at",
            "project",
            "is_outdated",
        )
        load_only = ("data_id",)

    data_id = fields.Integer()
    data_type = fields.Enum(constant_enum(DataType.geo_data), required=True)


class DataGeoCreationSchema(DataGeoSchema):
    """Schema for the creation of a new project geo data"""

    data_id = fields.Integer(required=True)


class ProjectGlobalDataSchema(ProjectDataBaseSchema):
    class Meta(ProjectDataBaseSchema.Meta):
        model = ProjectGlobalData
        fields = (
            "id",
            "data_type",
            "created_at",
            "modified_at",
            "last_update",
            "project",
            "description",
            "name",
            "data",
            "attributes",
            "input_data",
            "output_data",
            "used_input_attributes",
            "used_output_attributes",
            "data_id",
            "is_outdated",
        )
        dump_only = (
            "id",
            "data_type",
            "data",
            "used_input_attributes",
            "used_output_attributes",
            "last_update",
            "created_at",
            "modified_at",
            "project",
            "is_outdated",
        )
        load_only = ("data_id",)

    data_id = fields.Integer()
    data_type = fields.Enum(constant_enum(DataType.global_data), required=True)


class ProjectGlobalDataCreationSchema(ProjectGlobalDataSchema):
    """Schema for the creation of a new project global data"""

    data_id = fields.Integer(required=True)


class DataStreamSubStreamSchema(StreamGeoDataSchema):
    # Dummy schema necessary as nested OneOfSchema with only/exclude fields
    # breaks the OpenAPI specification (only one schema will be created when
    # nested, no matter the differences in excluded/only fields)
    pass


class DataStreamSchema(ProjectDataBaseSchema):
    class Meta(ProjectDataBaseSchema.Meta):
        model = DataStream
        fields = (
            "id",
            "data_type",
            "created_at",
            "modified_at",
            "last_update",
            "project",
            "description",
            "name",
            "data",
            "stream",
            "attributes",
            "input_data",
            "output_data",
            "used_input_attributes",
            "used_output_attributes",
            "stream_id",
            "classes",
            "start",
            "step",
            "stop",
            "resolution",
            "is_outdated",
        )
        dump_only = (
            "id",
            "data_type",
            "created_at",
            "modified_at",
            "last_update",
            "project",
            "data",
            "used_input_attributes",
            "used_output_attributes",
            "stream",
            "is_outdated",
        )
        load_only = ("stream_id",)

    stream_id = fields.Integer(required=False)
    data_type = fields.Enum(constant_enum(DataType.data_stream), required=True)
    stream = fields.Nested(
        DataStreamSubStreamSchema,
        only=("id", "type"),
    )
    classes = fields.List(fields.List(fields.Float), allow_none=True)


class DataStreamCreationSchema(DataStreamSchema):
    """Schema for the creation of a new project global data"""

    stream_id = fields.Integer(required=True)


class DataGeneratorSchema(ProjectDataBaseSchema):
    class Meta(ProjectDataBaseSchema.Meta):
        model = DataGenerator
        fields = (
            "id",
            "data_type",
            "created_at",
            "modified_at",
            "last_update",
            "project",
            "description",
            "name",
            "input_data",
            "output_data",
            "data",
            "processing_model",
            "attributes",
            "used_input_attributes",
            "used_output_attributes",
            "is_outdated",
            "proc_models_list",
        )
        dump_only = (
            "id",
            "data_type",
            "created_at",
            "modified_at",
            "last_update",
            "project",
            "data",
            "used_input_attributes",
            "used_output_attributes",
            "is_outdated",
            "proc_models_list",
        )

    data_type = fields.Enum(constant_enum(DataType.generator), required=True)
    # This field is used to create new processing models
    processing_model = fields.Nested(
        ProcessingModelCreationSchema(exclude=("data_generator",))
    )
    proc_models_list = fields.Nested(
        PrefModelBaseSchema,
        attribute="_processing_models",
        many=True,
        only=("id", "model_type", "name"),
    )


class DataGeneratorCreationSchema(DataGeneratorSchema):
    name = fields.String(required=True, allow_none=False)


# The inheritance from ProjectDataBaseSchema is necessary so fields.Pluck
# nesting this schema will work (they test field on ProjectDataSchema
# not on its type schemas)
class ProjectDataSchema(OneOfSchema, ProjectDataBaseSchema):
    type_field = "data_type"
    type_schemas = {
        DataType.geo_data.name: DataGeoSchema,
        DataType.generator.name: DataGeneratorSchema,
        DataType.global_data.name: ProjectGlobalDataSchema,
        DataType.data_stream.name: DataStreamSchema,
    }

    def get_obj_type(self, obj):
        return obj.data_type.name


class ProjectDataCreationSchema(
    OneOfSchemaWithType, ProjectDataBaseCreationSchema
):
    type_field = "data_type"
    type_schemas = {
        DataType.geo_data.name: DataGeoCreationSchema,
        DataType.generator.name: DataGeneratorCreationSchema,
        DataType.global_data.name: ProjectGlobalDataCreationSchema,
        DataType.data_stream.name: DataStreamCreationSchema,
    }

    def get_obj_type(self, obj):
        return obj.data_type.name


class ProjectDataGeneratorCreationSchema(
    OneOfSchemaWithType, ProjectDataBaseCreationSchema
):
    type_field = "data_type"
    type_schemas = {DataType.generator.name: DataGeneratorCreationSchema}

    def get_obj_type(self, obj):
        return obj.data_type.name



from functools import wraps

from flask import current_app as app
from flask import request
from jwt import DecodeError, ExpiredSignatureError, InvalidTokenError
from werkzeug.datastructures import Authorization

from ..exceptions import RequestError, UserTokenError
from ..models import User
from ..security import jwt_handler


def check_credentials(username, password):
    """Returns the user with the given credentials, or None"""
    user = User.get_by_username(username)
    if user is not None and user.check_password(password):
        return user
    return None


def is_jwt_valid(request_data):
    """Return True if the JWT token is valid, False otherwise"""
    token = jwt_handler.get_token_from_request(request_data)
    try:
        user_id = jwt_handler.get_user_id(token)
        return user_id is not None
    except InvalidTokenError:
        # JWT is not valid
        return False


def check_jwt_authentication(request_data):
    """Return the user corresponding to the JWT, or None"""
    token = jwt_handler.get_token_from_request(request_data)
    if token is None:
        raise RequestError("No authentication token provided", 401)
    try:
        return User.get_from_jwt(token)
    except UserTokenError as err:
        raise RequestError(str(err), 401)
    except ExpiredSignatureError:
        raise RequestError("Invalid JSON Web Token: the token is expired", 401)
    except DecodeError:
        raise RequestError(
            "Invalid JSON Web Token: the signature is not valid", 401
        )
    except InvalidTokenError:
        # the token is not valid
        # TODO: handle the expired tokens
        raise RequestError("Invalid JSON Web Token", 401)
    return None


def check_refresh_token(request_data):
    """Return the user if the refresh token is valid."""
    token = jwt_handler.get_token_from_request(request_data)
    if token is None:
        raise RequestError("No refresh token provided", 401)
    try:
        return User.get_from_refresh_token(token)
    except UserTokenError as err:
        raise RequestError(str(err), 401)
    except ExpiredSignatureError:
        raise RequestError(
            "Invalid refresh JSON Web Token: the token is expired", 401
        )
    except DecodeError:
        raise RequestError(
            "Invalid refresh JSON Web Token: the signature is not valid", 401
        )
    except InvalidTokenError:
        raise RequestError("Invalid refresh JSON Web Token", 401)
    return None


def check_basic_authentication(request_data):
    """Return the user corresponding to the Basic credentials, or None"""
    header = app.config.get("API_BASIC_AUTHENTICATION_HEADER", "Authorization")
    auth_header = request.headers.get(header, None)
    authorization = Authorization.from_header(auth_header)
    if authorization is not None:
        # valid authorization header for basic authentication
        username = authorization["username"]
        password = authorization["password"]
        user = check_credentials(username, password)
        if user is None:
            raise RequestError("Invalid credentials", 401)
        return user
    raise RequestError("No authorization header found", 401)


def refresh_token_required(fn):
    """Authenticate user with refresh token.

    Also add user to list of keyargs.

    :param fn:
    :return:
    """

    @wraps(fn)
    def wrapper(*args, **kwargs):
        user = check_refresh_token(request)
        return fn(*args, user=user, **kwargs)

    wrapper.__apispec__ = getattr(fn, "__apispec__", {})
    wrapper.__apispec__["security"] = wrapper.__apispec__.get("security", [])
    wrapper.__apispec__["security"].append({"refreshAuth": []})
    wrapper.__apispec__["responses"] = wrapper.__apispec__.get("responses", {})
    wrapper.__apispec__["responses"][401] = {
        # "schema": MessageSchema,
        "description": "Invalid or missing refresh token",
    }
    return wrapper


def basic_authenticated(fn):
    """Authenticate user with basic authentication.

    Also add user to list of keyargs.

    :param fn:
    :return:
    """

    @wraps(fn)
    def wrapper(*args, **kwargs):
        user = check_basic_authentication(request)
        return fn(*args, user=user, **kwargs)

    wrapper.__apispec__ = getattr(fn, "__apispec__", {})
    wrapper.__apispec__["security"] = wrapper.__apispec__.get("security", [])
    wrapper.__apispec__["security"].append({"basicAuth": []})
    wrapper.__apispec__["responses"] = wrapper.__apispec__.get("responses", {})
    wrapper.__apispec__["responses"][401] = {
        # "schema": MessageSchema,
        "description": "Error with basic authentification",
    }
    return wrapper


def token_required(fn):
    """wrapper to verify that the request contains a jwt token

    sends a 401 error if the token is missing or incorrect
    """

    @wraps(fn)
    def wrapper(*args, **kwargs):
        if is_jwt_valid(request):
            return fn(*args, **kwargs)
        else:
            raise RequestError("JSON Web Token invalid", 401)

    wrapper.__apispec__ = getattr(fn, "__apispec__", {})
    wrapper.__apispec__["security"] = wrapper.__apispec__.get("security", [])
    wrapper.__apispec__["security"].append({"bearerAuth": []})
    wrapper.__apispec__["responses"] = wrapper.__apispec__.get("responses", {})
    wrapper.__apispec__["responses"][401] = {
        # "schema": MessageSchema,
        "description": "Error with token authentification",
    }
    return wrapper



import enum
import numbers
from copy import copy, deepcopy
from datetime import datetime
from functools import total_ordering
from typing import Type

import simplejson as json
from geoalchemy2.shape import to_shape
from geoalchemy2.types import Geometry
from pandas import DataFrame
from shapely.geometry import mapping
from sqlalchemy.dialects.postgresql import ARRAY, JSONB
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.ext.hybrid import hybrid_property

from . import db
from .mixins import ModelMixin, TimestampMixin
from .permission import HasPermissions


class GeoDataType(enum.Enum):
    geo_data = 1
    generated_geo_data = 2
    global_data = 3
    wfs = 4
    wms = 5


class BaseData(HasPermissions, TimestampMixin, db.Model):
    type = db.Column(db.Enum(GeoDataType))  # inheritance discriminator
    name = db.Column(db.String, nullable=False)
    original_name = db.Column(db.String)
    description = db.Column(db.String)
    _properties_modified_at = db.Column("properties_modified_at", db.DateTime)

    attributes = db.relationship(
        "DataAttribute", back_populates="data", cascade="all,delete-orphan"
    )

    features = db.relationship(
        "Feature", back_populates="data", cascade="all,delete-orphan"
    )
    shares = db.relationship(
        "DataShare", cascade="all,delete-orphan", back_populates="data"
    )

    __mapper_args__ = {"polymorphic_on": type}

    def __init__(self, name, features=[], description=None, **kwargs):
        super().__init__(**kwargs)
        self.name = name
        self.features = features
        self.description = description

    def __repr__(self):
        return "<BaseData(name={})>".format(self.name)

    def __deepcopy__(self, memo):
        cls = type(self)
        other = cls(
            name=self.name,
        )
        memo[id(self)] = other
        return other

    def load_properties(self, force=False):
        """Load the model properties and save them to the database

        Args:
            force (bool): to force the loading of the properties
        """
        if (
            force
            or self.modified_at is None
            or self._properties_modified_at is None
            or self._properties_modified_at < self.modified_at
        ):
            for attribute in self.attributes:
                attribute.load_statistics()
            self._properties_modified_at = datetime.utcnow()
            self.save()


class BaseGeoData(BaseData):
    extent_filter = db.Column(Geometry)
    extent = db.Column(Geometry)

    @declared_attr
    def features(cls):
        return db.relationship(
            "GeoFeature", back_populates="data", cascade="all,delete-orphan"
        )

    __mapper_args__ = {"polymorphic_abstract": True}

    def __repr__(self):
        return "<BaseGeoData(name={})>".format(self.name)

    def __deepcopy__(self, memo):
        cls = type(self)
        other = cls(
            name=self.name,
        )
        memo[id(self)] = other
        other.features = [deepcopy(f, memo) for f in self.features]
        return other

    def as_geojson(self):
        features_geojson = [f.as_geojson() for f in self.features]
        collection = {
            "type": "FeatureCollection",
            "features": features_geojson,
        }
        return collection

    def load_extent(self):
        self.extent = (
            db.session.query(
                db.func.st_envelope(db.func.st_collect(GeoFeature.geom))
            )
            .filter_by(data_id=self.id)
            .one()[0]
        )

    def load_properties(self, force=False):
        """Load the model properties and save them to the database

        Args:
            force (bool): to force the loading of the properties
        """
        if (
            force
            or self.modified_at is None
            or self._properties_modified_at is None
            or self._properties_modified_at < self.modified_at
        ):
            self.load_extent()
            for attribute in self.attributes:
                attribute.load_statistics()
            self._properties_modified_at = datetime.utcnow()
            self.save()


class UploadableData:
    upload_user_id = db.Column(db.Integer, db.ForeignKey("user.id"))

    @declared_attr
    def _upload_user(cls):
        return db.relationship("User")

    @hybrid_property
    def upload_user(self):
        return self._upload_user

    @upload_user.setter
    def upload_user(self, user):
        """Make sure that the user has permissions on the data."""
        self._upload_user = user
        if user is not None:
            self.permissions.append(type(self).Permission(user=user))


class StreamGeoData:
    title = db.Column(db.String)
    url = db.Column(db.String, nullable=False)
    keywords = db.Column(ARRAY(db.String, dimensions=1))
    version = db.Column(db.String)


class WFSGeoData(UploadableData, StreamGeoData, BaseGeoData):
    __tablename__ = "wfs_geo_data"

    id = db.Column(db.Integer, db.ForeignKey("base_data.id"), primary_key=True)
    feature_type = db.Column(db.String, nullable=False)

    __mapper_args__ = {"polymorphic_identity": GeoDataType.wfs}

    def __init__(
        self,
        url: str,
        feature_type: str,
        name: str = None,
        original_name: str = None,
        is_public: bool = False,
        title: str = None,
        description: str = None,
        keywords: list = None,
        **kwargs,
    ):
        super().__init__(
            name=(name or feature_type),
            is_public=is_public,
            description=description,
            **kwargs,
        )
        self.url = url
        self.original_name = original_name
        self.title = title
        self.keywords = keywords if keywords else []
        self.feature_type = feature_type

    def __repr__(self):
        return "<WFSGeoData(name={})>".format(self.name)

    def __deepcopy__(self, memo):
        print("{}".format(self.name))
        memo[id(self)] = self
        return self


class WMSGeoData(UploadableData, StreamGeoData, BaseGeoData):
    __tablename__ = "wms_geo_data"

    id = db.Column(db.Integer, db.ForeignKey("base_data.id"), primary_key=True)
    classes = db.Column(ARRAY(db.Float, dimensions=2))
    start = db.Column(db.Float)
    step = db.Column(db.Float)
    stop = db.Column(db.Float)
    layer = db.Column(db.String, nullable=False)
    resolution = db.Column(db.Float)

    __mapper_args__ = {"polymorphic_identity": GeoDataType.wms}

    def __init__(
        self,
        url: str,
        layer: str,
        name: str = None,
        original_name: str = None,
        is_public: bool = False,
        title: str = None,
        description: str = None,
        keywords: list = None,
        classes: list[list[float]] = None,
        start: float = None,
        step: float = None,
        stop: float = None,
        **kwargs,
    ):
        super().__init__(
            name=(name or layer),
            is_public=is_public,
            description=description,
            **kwargs,
        )
        self.url = url
        self.original_name = original_name
        self.title = title
        self.keywords = keywords if keywords else []
        self.layer = layer
        self.classes = classes
        self.start = start
        self.step = step
        self.stop = stop

    def __repr__(self):
        return "<WMSGeoData(name={})>".format(self.name)

    def __deepcopy__(self, memo):
        print("{}".format(self.name))
        memo[id(self)] = self
        return self


class GlobalData(UploadableData, BaseData):
    __tablename__ = "global_data"

    id = db.Column(db.Integer, db.ForeignKey("base_data.id"), primary_key=True)
    feature_id = db.Column(db.Integer, db.ForeignKey("feature.id"))
    feature = db.relationship(
        "Feature",
        cascade="all",
        primaryjoin="global_data.c.feature_id == feature.c.id",
        post_update=True,
    )
    properties = db.relationship(
        "DataValue",
        cascade="all",
        primaryjoin=(
            "global_data.c.feature_id == foreign(data_value.c.feature_id)"
        ),
        post_update=True,
    )

    __mapper_args__ = {"polymorphic_identity": GeoDataType.global_data}

    def __init__(self, feature, **kwargs):
        super().__init__(**kwargs)
        self.feature = feature

    def __repr__(self):
        properties = {val.attribute.name: val.value for val in self.properties}
        return f"<GlobalData(name={self.name}, properties={properties})>"

    def __deepcopy__(self, memo):
        print("{}".format(self.name))
        memo[id(self)] = self
        return self

    def get_property(self, prop):
        """Returns the value of the property"""
        return self.feature.get_property(prop)

    def _choose_attribute_type(self, value) -> Type["DataAttribute"]:
        """Create new DataValue and DataAttribute based on value.

        :param prop: property name (attribute name)
        :param value:
        :return: created data value object (not persisted yet!)
        """
        match value:
            case int() | float():
                return DataAttributeQuantitative
            case _:
                return DataAttributeNominal

    def set_property(self, prop, value):
        """Set property value.

        Try to reuse existing attribute if type matches.

        :param prop: property name (attribute name)
        :param value: new property value
        """
        old_prop = None
        for p in self.properties:
            if p.attribute.name == prop:
                old_prop = p
                break
        attr_type = self._choose_attribute_type(value)
        if old_prop is None or not isinstance(old_prop.attribute, attr_type):
            attr = attr_type(name=prop, data=self)
            db.session.add(attr)
        else:
            attr = old_prop.attribute
            db.session.delete(old_prop)
        db.session.add(
            attr.get_value_class()(
                value=value, attribute=attr, feature=self.feature
            )
        )
        db.session.commit()


class GeoData(UploadableData, BaseGeoData):
    """Table for the geographical data loaded from files.

    Attributes:
        original_name (str): The name of the file.
        source_driver (str): Which format the original data used.
        upload_user_id (int): Reference to the user who uploaded the data.
        description (str): A description text.
    """

    __tablename__ = "geo_data"

    id = db.Column(db.Integer, db.ForeignKey("base_data.id"), primary_key=True)
    source_driver = db.Column(db.String)

    __mapper_args__ = {"polymorphic_identity": GeoDataType.geo_data}

    def __init__(
        self,
        *args,
        original_name=None,
        source_driver=None,
        is_public=False,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.original_name = original_name
        self.source_driver = source_driver
        self.is_public = is_public

    def __deepcopy__(self, memo):
        print("{}".format(self.name))
        memo[id(self)] = self
        return self

    def __repr__(self):
        return "<GeoData(name={}, source_driver={})>".format(
            self.name, self.source_driver
        )


class GeneratedGeoData(BaseGeoData):
    """Table for the geographical data generated by the processing model in the
    projects."""

    __mapper_args__ = {"polymorphic_identity": GeoDataType.generated_geo_data}

    project_data = db.relationship(
        "DynamicData",
        uselist=False,
        foreign_keys="DynamicData.data_id",
    )

    def __repr__(self):
        return "<GeneratedGeoDate(name={})>".format(self.name)

    def __deepcopy__(self, memo):
        other = super().__deepcopy__(memo)
        other.features = [deepcopy(f, memo) for f in self.features]
        other.project_data = deepcopy(self.project_data, memo)
        return other


class AttributeType(enum.Enum):
    quantitative = 1
    nominal = 2
    ordinal = 3


class DataAttribute(ModelMixin, db.Model):
    __tablename__ = "data_attribute"

    type = db.Column(db.Enum(AttributeType))
    name = db.Column(db.String, nullable=False)
    statistics = db.Column(JSONB)
    data_id = db.Column(
        db.Integer, db.ForeignKey("base_data.id"), nullable=False, index=True
    )

    data = db.relationship("BaseData", back_populates="attributes")
    values = db.relationship(
        "DataValue", back_populates="attribute", cascade="all,delete-orphan"
    )

    __mapper_args__ = {"polymorphic_on": type}

    def __init__(self, **kwargs):
        print("Attribute {} created".format(kwargs.get("name")))
        super().__init__(**kwargs)

    def __deepcopy__(self, memo):
        if self.data.type is GeoDataType.geo_data:
            memo[id(self)] = self
            return self
        else:
            cls = type(self)
            other = cls(
                name=self.name,
            )
            memo[id(self)] = other
            other.name = self.name
            other.data = deepcopy(
                self.data, memo
            )  # TODO: do not duplicate data
            other.values = [deepcopy(v, memo) for v in self.values]
            return other

    def new_copy(self):
        cls = self.__class__
        return cls(type=self.type, name=self.name)

    def load_statistics(self):
        """Load statistics about the attribute values.

        The stats are calculated using pandas' describe function.
        """
        values = [v.value for v in self.values]
        self.statistics = {"total": len(values)}
        df = DataFrame(values)
        if len(values) > 0 and len(df.columns) > 0:
            stats = df.describe(include="all")
            stats_dict = json.loads(stats[0].to_json())
            for key, value in stats_dict.items():
                if "%" in key:
                    self.statistics.setdefault("percentiles", {})[key] = value
                else:
                    self.statistics[key] = value

    def get_value_class(self):
        raise NotImplementedError

    def same_as(self, other) -> bool:
        """Check attribute is same as other.

        :param other: other data attribute or ``None``
        :return: ``True`` if they are the same, ``False`` otherwise

        .. warning:: This does not check values or statistics
        """
        return (
            isinstance(other, DataAttribute)
            and self.type == other.type
            and self.name == other.name
        )


class DataAttributeQuantitative(DataAttribute):
    __mapper_args__ = {"polymorphic_identity": AttributeType.quantitative}

    def get_value_class(self):
        return DataValueQuantitative


class DataAttributeNominal(DataAttribute):
    __mapper_args__ = {"polymorphic_identity": AttributeType.nominal}

    def get_value_class(self):
        return DataValueNominal


class DataAttributeOrdinal(DataAttribute):
    order = db.Column(JSONB)  # maybe not the best type?

    __mapper_args__ = {"polymorphic_identity": AttributeType.ordinal}

    def __deepcopy__(self, memo):
        other = super().__deepcopy__(memo)
        other.order = copy(self.order)
        return other

    def new_copy(self):
        obj = super().new_copy()
        obj.order = list(self.order)
        return obj

    def get_value_class(self):
        return DataValueOrdinal

    def same_as(self, other) -> bool:
        """Check attribute is same as other.

        :param other: other data attribute or ``None``
        :return: ``True`` if they are the same, ``False`` otherwise

        .. warning:: This does not check values or statistics
        """
        return super().same_as(other) and self.order == other.order


@total_ordering
class DataValue(ModelMixin, db.Model):
    __tablename__ = "data_value"

    type = db.Column(db.Enum(AttributeType))
    attribute_id = db.Column(
        db.Integer, db.ForeignKey("data_attribute.id"), index=True
    )
    feature_id = db.Column(
        db.Integer, db.ForeignKey("feature.id", ondelete="CASCADE"), index=True
    )
    value = db.Column(JSONB)

    __table_args__ = (db.UniqueConstraint("attribute_id", "feature_id"),)
    __mapper_args__ = {"polymorphic_on": type}

    attribute = db.relationship("DataAttribute", back_populates="values")
    feature = db.relationship("Feature", back_populates="properties")

    def __hash__(self):
        return hash((self.id, self.attribute_id, self.feature_id, self.value))

    def __lt__(self, other):
        raise NotImplementedError

    def __eq__(self, other):
        raise NotImplementedError

    def __deepcopy__(self, memo):
        other = type(self)()
        memo[id(self)] = other
        other.value = copy(self.value)
        other.attribute = deepcopy(self.attribute, memo)
        other.feature = deepcopy(self.feature, memo)
        return other


class DataValueQuantitative(DataValue):
    __mapper_args__ = {"polymorphic_identity": AttributeType.quantitative}

    def _check_type(self, other):
        if isinstance(other, DataValueQuantitative) is False:
            raise TypeError

    def __hash__(self):
        return super().__hash__()

    def __add__(self, other):
        if isinstance(other, numbers.Number):
            return self.__class__(value=self.value + other)
        elif isinstance(other, DataValueQuantitative):
            return self.__class__(value=self.value + other.value)
        else:
            return NotImplemented

    def __radd__(self, other):
        if isinstance(other, numbers.Number):
            return self.__class__(value=self.value + other)
        elif isinstance(other, DataValueQuantitative):
            return self.__class__(value=self.value + other.value)
        else:
            return NotImplemented

    def __truediv__(self, other):
        if isinstance(other, numbers.Number):
            return self.__class__(value=(self.value / other))
        elif isinstance(other, DataValueQuantitative):
            return self.__class__(value=(self.value / other.value))
        else:
            return NotImplemented

    def __lt__(self, other):
        """Check if self < other."""
        self._check_type(other)
        return self.value < other.value

    def __eq__(self, other):
        """Check if self == other."""
        self._check_type(other)
        return self.value == other.value


class DataValueNominal(DataValue):
    __mapper_args__ = {"polymorphic_identity": AttributeType.nominal}

    def _check_type(self, other):
        if isinstance(other, DataValueNominal) is False:
            raise TypeError

    def __hash__(self):
        return super().__hash__()

    def __lt__(self, other):
        """Check if self < other."""
        raise TypeError

    def __eq__(self, other):
        """Check if self == other."""
        self._check_type(other)
        return self.value == other.value


class DataValueOrdinal(DataValue):
    __mapper_args__ = {"polymorphic_identity": AttributeType.ordinal}

    def _check_type(self, other):
        if (
            isinstance(self.attribute, DataAttributeOrdinal) is False
            or self.attribute is not other.attribute
        ):
            raise TypeError

    def __hash__(self):
        return super().__hash__()

    def __lt__(self, other):
        """Check if self < other."""
        self._check_type(other)
        order = self.attribute.order
        return order.index(self.value) > order.index(other.value)

    def __eq__(self, other):
        """Check if self == other."""
        self._check_type(other)
        order = self.attribute.order
        return order.index(self.value) == order.index(other.value)


class FeatureType(enum.Enum):
    feature = 1
    geo_feature = 2


feature_input_assoc = db.Table(
    "feature_input_association",
    db.Model.metadata,
    db.Column(
        "feature_id",
        db.Integer,
        db.ForeignKey("feature.id"),
        primary_key=True,
    ),
    db.Column(
        "input_feature_id",
        db.Integer,
        db.ForeignKey("feature.id"),
        primary_key=True,
    ),
)


class Feature(ModelMixin, db.Model):
    """Table for the data features.

    Attributes:
        properties (dict): The properties or attributes of the feature.
        data_id (int): Reference to the data.
    """

    __tablename__ = "feature"

    id = db.Column(db.Integer, primary_key=True)
    type = db.Column(db.Enum(FeatureType))  # inheritance discriminator
    data_id = db.Column(
        db.Integer, db.ForeignKey("base_data.id"), nullable=False
    )
    execution_artifact = db.Column(JSONB)
    input_features = db.relationship(
        "Feature",
        secondary=feature_input_assoc,
        primaryjoin=(id == feature_input_assoc.c.feature_id),
        secondaryjoin=(id == feature_input_assoc.c.input_feature_id),
        backref="output_features",
        order_by=feature_input_assoc.c.input_feature_id,
    )

    data = db.relationship("BaseData", back_populates="features")
    properties = db.relationship(
        "DataValue", back_populates="feature", cascade="all,delete-orphan"
    )

    __mapper_args__ = {
        "polymorphic_on": type,
        "polymorphic_identity": FeatureType.feature,
    }

    def __init__(
        self,
        *args,
        properties=[],
        execution_artifact=None,
        input_features=[],
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.execution_artifact = execution_artifact or {}
        self.properties = properties
        self.input_features = input_features

    def __deepcopy__(self, memo):
        cls = type(self)
        other = cls(execution_artifact=copy(self.execution_artifact))
        memo[id(self)] = other
        other.data = deepcopy(self.data, memo)
        other.properties = [deepcopy(p, memo) for p in self.properties]
        return other

    def get_property(self, prop):
        """Returns the value of the property"""
        for p in self.properties:
            if p.attribute.name == prop:
                return p.value

    def get_input_data(self):
        """Returns the list of input data."""
        input_data = set()
        for feature in self.input_features:
            input_data.add(feature.data)
        return sorted(input_data, key=lambda d: d.id)

    def get_input_features(self, data_id):
        """Returns the input features belonging to provided data."""
        return [
            feature
            for feature in self.input_features
            if feature.data_id == data_id
        ]

    def get_output_data(self):
        """Returns the list of output data."""
        output_data = set()
        for feature in self.output_features:
            output_data.add(feature.data)
        return sorted(output_data, key=lambda d: d.id)

    def get_output_features(self, data_id):
        """Returns the output features belonging to provided data."""
        return [
            feature
            for feature in self.output_features
            if feature.data_id == data_id
        ]

    def explain(self):
        """Return explanation for feature values.

        This is empty except for generated geo data features.

        :return:
        """
        if self.data.type != GeoDataType.generated_geo_data:
            return {}
        return self.data.project_data.explain(self)


class GeoFeature(Feature):
    """Table for the geo-data features.

    Attributes:
        geom: The geometry of the feature.
        properties (dict): The properties or attributes of the feature.
        data_id (int): Reference to the geo-data.
    """

    __tablename__ = "geo_feature"

    id = db.Column(db.Integer, db.ForeignKey("feature.id"), primary_key=True)
    data = db.relationship("BaseGeoData", back_populates="features")
    geom = db.Column(Geometry)

    __mapper_args__ = {"polymorphic_identity": FeatureType.geo_feature}

    def __init__(self, geom, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.geom = geom

    def as_geojson(self):
        """Return the feature as a GeoJSON data."""
        geometry = mapping(to_shape(self.geom))
        geojson = {
            "type": "Feature",
            "geometry": geometry,
            "properties": {p.attribute.name: p.value for p in self.properties},
        }
        return geojson

    def __deepcopy__(self, memo):
        cls = type(self)
        other = cls(
            geom=copy(self.geom),
            execution_artifact=copy(self.execution_artifact),
        )
        memo[id(self)] = other
        other.data = deepcopy(self.data, memo)
        other.properties = [deepcopy(p, memo) for p in self.properties]
        return other



import io

import simplejson as json
from flask import jsonify, request, send_file
from flask_restful import Resource, reqparse
from sqlalchemy.exc import IntegrityError

from ..exceptions import (
    CapabilitiesXMLParsingError,
    ExternalRequestError,
    RequestError,
    UploadError,
)
from ..models import (
    BaseData,
    DataAttribute,
    Feature,
    GeoData,
    GeoDataType,
    GlobalData,
    PermissionAbility,
    StreamGeoData,
    db,
)
from ..schemas import (
    BaseDataFeaturesSchema,
    DataAttributeGetResponse,
    DataAttributeSchema,
    DataAttributeSchemaWithValues,
    DataAttributeWithValuesGetResponse,
    DataSchema,
    DataTypeSetSchema,
    FeatureResponseSchema,
    FeatureSchema,
    GeoDataAccessResponseSchema,
    GeoDataAccessSchema,
    GeoDataGetResponse,
    GeoDataListResponseSchema,
    GeoDataSchema,
    GeoOnlyDataSchema,
    GlobalDataAccessResponseSchema,
    GlobalDataAccessSchema,
    GlobalDataCreationSchema,
    GlobalDataSchema,
    MessageSchema,
    StreamGeoDataAccessResponseSchema,
    StreamGeoDataAccessSchema,
    StreamGeoDataListResponseSchema,
    StreamGeoDataResponseSchema,
    StreamGeoDataSchema,
    WFSGeoDataCreationSchema,
    WMSGeoDataCreationSchema,
)
from ..services import (
    data_service,
    geo_data_service,
    geo_data_stream_service,
    get_global_data_service,
)
from ..services.auth_service import check_jwt_authentication, token_required
from ..services.geo_data_loading_service import (
    create_global_data,
    get_geo_data_download,
    get_temporary_directory,
    load_from_wfs,
    load_from_wms,
    load_raster_from_files,
    load_vector_from_files,
    load_wfs,
    load_wms,
    post_load_normalization,
)
from ..services.permission_service import has_ability
from .utils import (
    dump_data,
    get_json_content,
    with_query_arg,
    with_request_body,
    with_response,
)


class DataListAPI(Resource):
    def parse_types(self, args) -> set[str]:
        res = set()
        types = args.get("type", None)
        all_types_names = {_t.name for _t in GeoDataType}
        if types is None or types == [""]:
            return all_types_names
        for _type in types:
            if _type in all_types_names:
                res.add(GeoDataType[_type].name)
        return res

    @token_required
    @with_query_arg(
        "type",
        schema=DataTypeSetSchema,
        required=False,
        description="The types of data to return",
    )
    @with_response(
        status=200,
        schema=DataSchema(many=True),
        description="List of all data the user has access to",
    )
    def get(self):
        """
        Get the list of all the data for the user.

        :reqheader Authorization:
            JSON Web Token with Bearer scheme
            (``Authorization: Bearer <token>``).
        :status 403: No user logged in.

        .. :quickref: Data; Get the list of accessible data
        """
        parser = reqparse.RequestParser()
        parser.add_argument("type", type=str, location="args", action="append")
        args = parser.parse_args()
        types = self.parse_types(args)
        user = check_jwt_authentication(request)
        if user is None:
            raise PermissionError("No user logged in")
        data_list = data_service.get_all_authorized_data(user, types)
        return jsonify(DataSchema(many=True).dump(data_list))


class DataAPI(Resource):
    @token_required
    @with_response(status=200, schema=DataSchema, description="Data details")
    def get(self, data_id):
        """
        Get a data from its id.

        :param int data_id: The id of the data.
        :reqheader Authorization:
            JSON Web Token with Bearer scheme
            (``Authorization: Bearer <token>``).
        :status 403: The user is not allowed to access this data.
        :status 404: The data doesn't exist.

        .. :quickref: Data; Get the data details
        """
        data = data_service.get_if_authorized(data_id)
        return jsonify(DataSchema().dump(data))


class GlobalDataListAPI(Resource):
    @token_required
    @with_response(
        status=200,
        schema=GlobalDataSchema(many=True),
        description="List of all global data the user has access to",
    )
    def get(self):
        """
        Get the list of all the global data for the user.

        :reqheader Authorization:
            JSON Web Token with Bearer scheme
            (``Authorization: Bearer <token>``).
        :status 403: No user logged in.

        .. :quickref: Globaldata; Get the list of accessible global data
        """
        user = check_jwt_authentication(request)
        if user is None:
            raise PermissionError("No user logged in")
        data_list = get_global_data_service.get_all_authorized_data(user)
        return jsonify(GlobalDataSchema(many=True).dump(data_list))

    @token_required
    @with_request_body(
        schema=GlobalDataCreationSchema, description="Global data details"
    )
    @with_response(
        status=200,
        schema=GlobalDataSchema,
        description="Created global data details",
    )
    @has_ability(PermissionAbility.create_geo_data)
    def post(self):
        """
        Create an existing global data.

        :reqheader Authorization:
            JSON Web Token with Bearer scheme
            (``Authorization: Bearer <token>``).
        :status 403: The user is not allowed to create global data.

        .. :quickref: Globaldata; Create a global data
        """
        user = check_jwt_authentication(request)
        content = get_json_content()
        schema = GlobalDataSchema()
        data_json = GlobalDataCreationSchema().load(content)
        data = create_global_data(**data_json)
        data.upload_user = user
        data.update()
        return jsonify(schema.dump(data))


class GlobalDataAPI(Resource):
    @token_required
    @with_response(
        status=200, schema=GlobalDataSchema, description="Global data details"
    )
    def get(self, data_id):
        """
        Get a global data from its id.

        :param int data_id: The id of the global data.
        :reqheader Authorization:
            JSON Web Token with Bearer scheme
            (``Authorization: Bearer <token>``).
        :status 403: The user is not allowed to access this global data.
        :status 404: The global data doesn't exist.

        .. :quickref: Globaldata; Get the global data details
        """
        data = get_global_data_service.get_if_authorized(data_id)
        return jsonify(GlobalDataSchema().dump(data))

    @token_required
    @with_response(
        status=200,
        schema=MessageSchema,
        description="Global data successfully deleted",
    )
    def delete(self, data_id):
        """
        Delete a global data.

        :param int data_id: The id of the global -data.
        :reqheader Authorization:
            JSON Web Token with Bearer scheme
            (``Authorization: Bearer <token>``).
        :status 400:
            The global data is used as input in other data and cannot be
            deleted.
        :status 403: The user is not allowed to delete the global data.
        :status 404: The global data doesn't exist.

        .. :quickref: Globaldata; Delete a global data
        """
        data = get_global_data_service.get_if_authorized(data_id)
        try:
            data.delete()
        except IntegrityError:
            raise RequestError(
                "The global data is referenced by other objects and cannot "
                "be deleted",
                400,
            )
        return jsonify(message="global data {} deleted".format(data_id))

    @token_required
    @with_request_body(
        schema=GlobalDataSchema, description="New global data details"
    )
    @with_response(
        status=200,
        schema=GlobalDataSchema,
        description="Global data updated details",
    )
    def put(self, data_id):
        """
        Update an existing global data.

        :param int data_id: The id of the global data.
        :reqheader Authorization:
            JSON Web Token with Bearer scheme
            (``Authorization: Bearer <token>``).
        :status 403: The user is not allowed to modify the global data.
        :status 404: The global data doesn't exist.

        .. :quickref: Globaldata; Update a global data
        """
        content = get_json_content()
        schema = GlobalDataSchema()
        data = get_global_data_service.get_if_authorized(data_id)
        if "properties" in content:
            for prop in data.properties:
                db.session.delete(prop)
        data = schema.load(content, instance=data)
        if "properties" in content:
            properties = {prop.attribute: prop for prop in data.properties}
            for attr in data.attributes:
                prop = properties.get(attr)
                if prop:
                    db.session.add(attr)
                    db.session.add(prop)
                else:
                    db.session.delete(attr)
        data.update()
        return jsonify(schema.dump(data))


class GlobalDataAccessAPI(Resource):
    @token_required
    @with_response(
        status=200,
        schema=GlobalDataAccessResponseSchema,
        description="Global data permissions",
    )
    def get(self, data_id):
        """
        Return the list of permissions for the global data.

        The user must be the owner of the global data.

        :param int data_id: The id of the data.
        :reqheader Authorization:
            JSON Web Token with Bearer scheme
            (``Authorization: Bearer <token>``).
        :status 403: The user is not the owner of the data.
        :status 404: The global data doesn't exist.

        .. :quickref: Globaldata; Get the global data permissions
        """
        data = GlobalData.get_by_id(data_id)
        if data is None:
            raise RequestError("The global data does not exist", 404)
        user = check_jwt_authentication(request)
        if data.upload_user is None or data.upload_user is not user:
            raise PermissionError(
                "The user is not the owner of the global data"
            )
        schema = GlobalDataAccessSchema()
        return dump_data(schema, access=data)

    @token_required
    @with_request_body(
        schema=GlobalDataAccessSchema,
        description="New global data permissions",
    )
    @with_response(
        status=200,
        schema=GlobalDataAccessResponseSchema,
        description="Global data updated permissions",
    )
    def put(self, data_id):
        """
        Update the list of permissions for the global data.

        The user must be the owner of the global data.

        :param int data_id: The id of the data.
        :reqheader Authorization:
            JSON Web Token with Bearer scheme
            (``Authorization: Bearer <token>``).
        :status 403: The user is not the owner of the data.
        :status 404: The global data doesn't exist.

        .. :quickref: Globaldata; Modify the global data permissions
        """
        data = GlobalData.get_by_id(data_id)
        if data is None:
            raise RequestError("The global data does not exist", 404)
        user = check_jwt_authentication(request)
        if data.upload_user is None or data.upload_user is not user:
            raise PermissionError(
                "The user is not the owner of the global data"
            )
        schema = GlobalDataAccessSchema()
        content = get_json_content()
        data = schema.load(content, instance=data)
        if not any([p.user is data.upload_user for p in data.permissions]):
            data.permissions.append(
                GlobalData.Permission(user=data.upload_user)
            )
        data.update()
        return dump_data(schema, access=data)


class GeoDataAPI(Resource):
    @token_required
    @with_response(
        status=200, schema=GeoDataGetResponse, description="Geodata details"
    )
    def get(self, geo_data_id):
        """
        Get a geo-data from its id.

        :param int geo_data_id: The id of the geo-data.
        :reqheader Authorization:
            JSON Web Token with Bearer scheme
            (``Authorization: Bearer <token>``).
        :status 403: The user is not allowed to access this geo-data.
        :status 404: The geo-data doesn't exist.

        .. :quickref: Geodata; Get the geodata details
        """
        geo_data = geo_data_service.get_if_authorized(geo_data_id)
        return dump_data(GeoOnlyDataSchema(), geodata=geo_data)

    @token_required
    @with_response(
        status=200,
        schema=MessageSchema,
        description="Geodata successfully deleted",
    )
    def delete(self, geo_data_id):
        """
        Delete a geodata and its features.

        :param int geo_data_id: The id of the geo-data.
        :reqheader Authorization:
            JSON Web Token with Bearer scheme
            (``Authorization: Bearer <token>``).
        :status 400:
            The geo-data is used as input in other data and cannot be deleted.
        :status 403: The user is not allowed to delete the geo-data.
        :status 404: The geo-data doesn't exist.

        .. :quickref: Geodata; Delete a geodata
        """
        geo_data = geo_data_service.get_if_authorized(geo_data_id)
        try:
            geo_data.delete()
        except IntegrityError:
            raise RequestError(
                "The geo data is referenced by other objects and cannot "
                "be deleted",
                400,
            )
        return jsonify(message="geo-data {} deleted".format(geo_data_id))

    @token_required
    @with_request_body(schema=GeoDataSchema, description="New geodata details")
    @with_response(
        status=200,
        schema=GeoDataGetResponse,
        description="Geodata updated details",
    )
    def put(self, geo_data_id):
        """
        Update an existing geo-data.

        :param int geo_data_id: The id of the geo-data.
        :reqheader Authorization:
            JSON Web Token with Bearer scheme
            (``Authorization: Bearer <token>``).
        :status 403: The user is not allowed to modify the geo-data.
        :status 404: The geo-data doesn't exist.

        .. :quickref: Geodata; Update a geodata
        """
        content = get_json_content()
        schema = GeoDataSchema()
        geo_data = geo_data_service.get_if_authorized(geo_data_id)
        data = schema.load(content, instance=geo_data)
        data.update()
        return dump_data(schema, geodata=data)


class StreamGeoDataAPI(Resource):
    @token_required
    @with_response(
        status=200,
        schema=StreamGeoDataResponseSchema,
        description="Geodata stream details",
    )
    def get(self, stream_geo_data_id):
        """
        Get a geo-data stream from its id.

        :param int stream_geo_data_id: The id of the geo-data stream.
        :reqheader Authorization:
            JSON Web Token with Bearer scheme
            (``Authorization: Bearer <token>``).
        :status 403: The user is not allowed to access this geo-data stream.
        :status 404: The geo-data stream doesn't exist.

        .. :quickref: StreamGeoData; Get the geodata stream details
        """
        stream = geo_data_stream_service.get_if_authorized(stream_geo_data_id)
        return dump_data(StreamGeoDataSchema(), stream=stream)

    @token_required
    @with_response(
        status=200,
        schema=MessageSchema,
        description="Geodata stream successfully deleted",
    )
    def delete(self, stream_geo_data_id):
        """
        Delete a geodata stream.

        :param int stream_geo_data_id: The id of the geo-data stream.
        :reqheader Authorization:
            JSON Web Token with Bearer scheme
            (``Authorization: Bearer <token>``).
        :status 403: The user is not allowed to delete the geo-data stream.
        :status 404: The geo-data stream doesn't exist.

        .. :quickref: StreamGeoData; Delete a geodata stream
        """
        stream = geo_data_stream_service.get_if_authorized(stream_geo_data_id)
        try:
            stream.delete()
        except IntegrityError:
            # This should not happen as stream is detached gracefully
            # beforehand
            raise RequestError(
                "The geo data stream is referenced by other objects and cannot"
                " be deleted",
                500,
            )
        return jsonify(
            message="geo-data stream {} deleted".format(stream_geo_data_id)
        )

    @token_required
    @with_request_body(
        schema=StreamGeoDataSchema, description="New geodata stream details"
    )
    @with_response(
        status=200,
        schema=StreamGeoDataResponseSchema,
        description="Geodata stream updated details",
    )
    def put(self, stream_geo_data_id):
        """
        Update an existing geo-data stream.

        :param int stream_geo_data_id: The id of the geo-data stream.
        :reqheader Authorization:
            JSON Web Token with Bearer scheme
            (``Authorization: Bearer <token>``).
        :status 403: The user is not allowed to modify the geo-data stream.
        :status 404: The geo-data stream doesn't exist.

        .. :quickref: StreamGeoData; Update a geodata stream
        """
        content = get_json_content()
        stream = geo_data_stream_service.get_if_authorized(stream_geo_data_id)
        schema = StreamGeoDataSchema.type_schemas[stream.type.name]()
        stream = schema.load(content, instance=stream)
        stream.update()
        return dump_data(schema, stream=stream)


class StreamGeoDataUpdateAPI(Resource):
    @token_required
    @with_response(
        status=200,
        schema=StreamGeoDataResponseSchema,
        description="Geodata stream updated details",
    )
    def post(self, stream_geo_data_id):
        """
        Update features preview of an existing geo-data stream.

        :param int stream_geo_data_id: The id of the geo-data stream.
        :reqheader Authorization:
            JSON Web Token with Bearer scheme
            (``Authorization: Bearer <token>``).
        :status 400: The stream could not be retrieved with given parameters.
        :status 403: The user is not allowed to modify the geo-data stream.
        :status 404: The geo-data stream doesn't exist.

        .. :quickref: StreamGeoData; Update features preview of a geodata stream
        """  # noqa: E501
        stream = geo_data_stream_service.get_if_authorized(stream_geo_data_id)
        schema = StreamGeoDataSchema.type_schemas[stream.type.name]()
        try:
            if stream.type == GeoDataType.wfs:
                features = load_from_wfs(stream, stream.attributes)
            else:
                attribute = None
                if len(stream.attributes) == 1:
                    attribute = stream.attributes[0]
                features = load_from_wms(
                    stream,
                    old_attribute=attribute,
                )
        except (
            CapabilitiesXMLParsingError,
            ExternalRequestError,
            KeyError,
        ) as exc:
            raise RequestError(str(exc), 400)
        # List the data attributes
        attrs = []
        for f in features:
            for p in f.properties:
                if p.attribute not in attrs:
                    attrs.append(p.attribute)
        stream.features = features
        stream.extent = None
        stream.attributes = attrs
        stream.update()
        post_load_normalization(stream)
        return dump_data(schema, stream=stream)


class DataAttributeAPI(Resource):
    @token_required
    @with_response(
        status=200,
        schema=[DataAttributeGetResponse, DataAttributeWithValuesGetResponse],
        description="Details of an attribute",
    )
    @with_query_arg(
        "values",
        bool,
        required=False,
        description="Return the values of the attribute. Default: false",
    )
    def get(self, attribute_id):
        """
        Get the details of an attribute of a geo-data.

        The details contain the name of the attribute,
        its type and some statistics about its values.

        :param int attribute_id: The id of the data attribute.
        :query bool values: Return the values of the attribute. Default: False.
        :reqheader Authorization:
            JSON Web Token with Bearer scheme
            (``Authorization: Bearer <token>``).
        :status 403: The user is not allowed to access the geo-data attribute.
        :status 404: The geo-data attribute doesn't exist.

        .. :quickref: Geodata; Get the details of an attribute
        """
        # TODO: Should verify the user has permission to view the data
        parser = reqparse.RequestParser()
        parser.add_argument("values", type=bool, location="args")
        args = parser.parse_args()
        attribute = DataAttribute.get_by_id(attribute_id)
        geo_data_service.check_permission(attribute.data)
        if args.get("values", False):
            schema = DataAttributeSchemaWithValues()
        else:
            schema = DataAttributeSchema()
        return dump_data(schema, attribute=attribute)


class DataDownloadAPI(Resource):
    @token_required
    @with_query_arg(
        "outputFormat",
        str,
        required=False,
        description=(
            "Desired output format (leave blank for zip shapefile/json)"
        ),
    )
    @with_response(
        status=200,
        description="Data in specified file format",
        content={
            "application/octet-stream": {"type": "string", "format": "binary"}
        },
    )
    def get(self, data_id):
        """
        Return the data in desired format.

        :param int data_id: The id of the data to download.
        :reqheader Authorization:
            JSON Web Token with Bearer scheme
            (``Authorization: Bearer <token>``).
        :status 403: The user is not allowed to access the data.
        :status 404: The data doesn't exist.
        :status 415: Output format is not supported.
        :status 500:
            There was an error with the geometry while creating the file.

        .. :quickref: Data; Download a data as a file
        """
        parser = reqparse.RequestParser()
        parser.add_argument("outputFormat", type=str, location="args")
        args = parser.parse_args()
        geo_data = data_service.get_if_authorized(data_id)
        with get_temporary_directory() as temp_dir:
            filename, file_object = get_geo_data_download(
                geo_data, temp_dir, args.get("outputFormat")
            )
            return send_file(
                file_object, as_attachment=True, download_name=filename
            )


class GeoDataUploadAPI(Resource):
    @token_required
    @with_request_body(
        description="Geodata as zipped/filelist shapefile or other format",
        content={
            "multipart/form-data": {
                "schema": {
                    "type": "object",
                    "properties": {
                        "file": {
                            "type": "array",
                            "items": {"type": "string", "format": "binary"},
                        }
                    },
                }
            }
        },
    )
    @with_response(
        status=200,
        schema=GeoDataGetResponse,
        description="Uploaded geodata details",
    )
    @has_ability(PermissionAbility.create_geo_data)
    def post(self):
        """
        Create a new geo-data from the uploaded files.

        The uploaded file can either be a .zip file containing all the required
        shapefile files, the list of shapefile files or a single file
        containing vector geodata.

        :form file: The list of files.
        :reqheader Authorization:
            JSON Web Token with Bearer scheme
            (``Authorization: Bearer <token>``).
        :status 403: The user is not allowed to create geo-data.
        :status 415: There is an error in the uploaded file.

        .. :quickref: Geodata; Upload a vector geodata
        """
        if "file" in request.files:
            user = check_jwt_authentication(request)
            files_list = request.files.getlist("file")
            try:
                geo_data = load_vector_from_files(files_list)
            except UploadError as err:
                raise RequestError(str(err), 415)
            post_load_normalization(geo_data)
            geo_data.upload_user = user
            geo_data.update()
            return dump_data(GeoDataSchema(), geodata=geo_data)


class WFSGeoDataUploadAPI(Resource):
    @token_required
    @with_request_body(
        schema=WFSGeoDataCreationSchema,
        description="Parameters necessary to find the WFS stream online",
    )
    @with_response(
        status=200,
        schema=StreamGeoDataResponseSchema,
        description="Uploaded WFS geodata stream details",
    )
    @has_ability(PermissionAbility.create_geo_data)
    def post(self):
        """
        Create a new geo-data stream.

        :reqheader Authorization:
            JSON Web Token with Bearer scheme
            (``Authorization: Bearer <token>``).
        :status 400: The stream could not be retrieved with given parameters.
        :status 403: The user is not allowed to create geo-data stream.

        .. :quickref: StreamGeoData; Upload a WFS geo data stream
        """
        user = check_jwt_authentication(request)
        content = get_json_content()
        stream = WFSGeoDataCreationSchema().load(content)
        try:
            stream = load_wfs(stream)
        except (
            CapabilitiesXMLParsingError,
            ExternalRequestError,
            KeyError,
        ) as exc:
            raise RequestError(str(exc), 400)
        stream.upload_user = user
        stream.create()
        post_load_normalization(stream)
        return dump_data(StreamGeoDataSchema(), stream=stream)


class WMSGeoDataUploadAPI(Resource):
    @token_required
    @with_request_body(
        schema=WMSGeoDataCreationSchema,
        description="Parameters necessary to find the WMS stream online",
    )
    @with_response(
        status=200,
        schema=StreamGeoDataResponseSchema,
        description="Uploaded WMS geodata stream details",
    )
    @has_ability(PermissionAbility.create_geo_data)
    def post(self):
        """
        Create a new geo-data stream.

        :reqheader Authorization:
            JSON Web Token with Bearer scheme
            (``Authorization: Bearer <token>``).
        :status 400: The stream could not be retrieved with given parameters.
        :status 403: The user is not allowed to create geo-data stream.

        .. :quickref: StreamGeoData; Upload a WMS geo data stream
        """
        user = check_jwt_authentication(request)
        content = get_json_content()
        stream = WMSGeoDataCreationSchema().load(content)
        try:
            stream = load_wms(stream)
        except (
            CapabilitiesXMLParsingError,
            ExternalRequestError,
            KeyError,
        ) as exc:
            raise RequestError(str(exc), 400)
        stream.upload_user = user
        stream.create()
        post_load_normalization(stream)
        return dump_data(StreamGeoDataSchema(), stream=stream)


class GeoDataUploadRasterAPI(Resource):
    @token_required
    @with_request_body(
        description="Geodata as a raster image",
        content={
            "multipart/form-data": {
                "schema": {
                    "type": "object",
                    "properties": {
                        "file": {
                            "type": "array",
                            "items": {"type": "string", "format": "binary"},
                        },
                        "classes": {
                            "type": "array",
                            "items": {
                                "type": "array",
                                "items": {"type": "number"},
                            },
                        },
                        "start": {"type": "number"},
                        "step": {"type": "number"},
                        "stop": {"type": "number"},
                    },
                }
            }
        },
    )
    @with_response(
        status=200,
        schema=GeoDataGetResponse,
        description="Uploaded geodata details",
    )
    @has_ability(PermissionAbility.create_geo_data)
    def post(self):
        """
        Create a new vectorized geo-data from the uploaded raster file.

        The raster data is reclassified using the classes given,
        to group multiple values together
        and transform the raster file in vector geometries.

        For example, classes of ``[[0, 10, 5], [10, 100, 50]]``
        would assign the value 5 to all the raster cells in [0, 10[,
        the value 50 to all the cells in [10, 100[,
        and 0 or 100 to the cells outside that range.

        :form file: The raster file.
        :form classes:
            The classes to reclassify the raster values
            (list ``[[low, high, value], ...]``).
        :form start:
            The starting value for classes creation
            (if `classes` not provided).
        :form step:
            The step value for classes creation (if `classes` not provided).
        :form stop:
            The stop value for classes creation (if `classes` not provided).
        :reqheader Authorization:
            JSON Web Token with Bearer scheme
            (``Authorization: Bearer <token>``).
        :status 403: The user is not allowed to create geo-data.
        :status 415: There is an error in the uploaded file.

        .. :quickref: Geodata; Upload a raster geodata
        """
        if "file" in request.files:
            user = check_jwt_authentication(request)
            files_list = request.files.getlist("file")
            _classes_str = request.form.get("classes", "")
            _start_str = request.form.get("start", "")
            _step_str = request.form.get("step", "")
            _stop_str = request.form.get("stop", "")
            classes = (
                None if len(_classes_str) == 0 else json.loads(_classes_str)
            )  # parse the classes as json
            start = (
                None if len(_start_str) == 0 else json.loads(_start_str)
            )  # parse the start as json
            step = (
                None if len(_step_str) == 0 else json.loads(_step_str)
            )  # parse the step as json
            stop = (
                None if len(_stop_str) == 0 else json.loads(_stop_str)
            )  # parse the stop as json
            try:
                geo_data = load_raster_from_files(
                    files_list,
                    classes=classes,
                    start=start,
                    step=step,
                    stop=stop,
                )
            except UploadError as err:
                raise RequestError(str(err), 415)
            post_load_normalization(geo_data)
            geo_data.upload_user = user
            geo_data.update()
        return dump_data(GeoDataSchema(), geodata=geo_data)


class GeoDataAccessAPI(Resource):
    @token_required
    @with_response(
        status=200,
        schema=GeoDataAccessResponseSchema,
        description="Geodata permissions",
    )
    def get(self, geo_data_id):
        """
        Return the list of permissions for the geo-data.

        The user must be the owner of the geodata.

        :param int geo_data_id: The id of the data.
        :reqheader Authorization:
            JSON Web Token with Bearer scheme
            (``Authorization: Bearer <token>``).
        :status 403: The user is not the owner of the data.
        :status 404: The geo-data doesn't exist.

        .. :quickref: Geodata; Get the geodata permissions
        """
        geo_data = GeoData.get_by_id(geo_data_id)
        if geo_data is None:
            raise RequestError("The geo data does not exist", 404)
        user = check_jwt_authentication(request)
        if geo_data.upload_user is None or geo_data.upload_user is not user:
            raise PermissionError("The user is not the owner of the geo-data")
        schema = GeoDataAccessSchema()
        return dump_data(schema, access=geo_data)

    @token_required
    @with_request_body(
        schema=GeoDataAccessSchema, description="New geodata permissions"
    )
    @with_response(
        status=200,
        schema=GeoDataAccessResponseSchema,
        description="Geodata updated permissions",
    )
    def put(self, geo_data_id):
        """
        Update the list of permissions for the geo-data.

        The user must be the owner of the geodata.

        :param int geo_data_id: The id of the data.
        :reqheader Authorization:
            JSON Web Token with Bearer scheme
            (``Authorization: Bearer <token>``).
        :status 403: The user is not the owner of the data.
        :status 404: The geo-data doesn't exist.

        .. :quickref: Geodata; Modify the geodata permissions
        """
        geo_data = GeoData.get_by_id(geo_data_id)
        if geo_data is None:
            raise RequestError("The geo data does not exist", 404)
        user = check_jwt_authentication(request)
        if geo_data.upload_user is None or geo_data.upload_user is not user:
            raise PermissionError("The user is not the owner of the geo-data")
        schema = GeoDataAccessSchema()
        content = get_json_content()
        geo_data = schema.load(content, instance=geo_data)
        if not any(
            [p.user is geo_data.upload_user for p in geo_data.permissions]
        ):
            geo_data.permissions.append(
                GeoData.Permission(user=geo_data.upload_user)
            )
        geo_data.update()
        return dump_data(schema, access=geo_data)


class GeoDataListAPI(Resource):
    @token_required
    @with_response(
        status=200,
        schema=GeoDataListResponseSchema,
        description="List of all geodata the user has access to",
    )
    def get(self):
        """
        Get the list of all the geo-data for the user.

        :reqheader Authorization:
            JSON Web Token with Bearer scheme
            (``Authorization: Bearer <token>``).
        :status 403: No user logged in.

        .. :quickref: Geodata; Get the list of accessible geodata
        """
        user = check_jwt_authentication(request)
        if user is None:
            raise PermissionError("No user logged in")
        geo_data_list = geo_data_service.get_all_authorized_geo_data(user)
        return dump_data(GeoDataSchema(many=True), geodata=geo_data_list)


class StreamGeoDataAccessAPI(Resource):
    @token_required
    @with_response(
        status=200,
        schema=StreamGeoDataAccessResponseSchema,
        description="Geodata stream permissions",
    )
    def get(self, stream_geo_data_id):
        """
        Return the list of permissions for the geo-data stream.

        The user must be the owner of the geodata stream.

        :param int stream_geo_data_id: The id of the data.
        :reqheader Authorization:
            JSON Web Token with Bearer scheme
            (``Authorization: Bearer <token>``).
        :status 403: The user is not the owner of the geodata stream.
        :status 404: The geo-data stream doesn't exist.

        .. :quickref: StreamGeoData; Get the geodata stream permissions
        """
        stream = BaseData.get_by_id(stream_geo_data_id)
        user = check_jwt_authentication(request)
        if not isinstance(stream, StreamGeoData):
            raise RequestError("The geo data stream does not exist", 404)
        if stream.upload_user is not user:
            raise PermissionError(
                "The user is not the owner of the geo-data stream"
            )
        schema = StreamGeoDataAccessSchema()
        return dump_data(schema, access=stream)

    @token_required
    @with_request_body(
        schema=StreamGeoDataAccessSchema,
        description="New geodata stream permissions",
    )
    @with_response(
        status=200,
        schema=StreamGeoDataAccessResponseSchema,
        description="Geodata stream updated permissions",
    )
    def put(self, stream_geo_data_id):
        """
        Update the list of permissions for the geo-data.

        The user must be the owner of the geodata stream.

        :param int stream_geo_data_id: The id of the data.
        :reqheader Authorization:
            JSON Web Token with Bearer scheme
            (``Authorization: Bearer <token>``).
        :status 403: The user is not the owner of the geodata stream.
        :status 404: The geo-data stream doesn't exist.

        .. :quickref: StreamGeoData; Modify the geodata stream permissions
        """
        stream = BaseData.get_by_id(stream_geo_data_id)
        if not isinstance(stream, StreamGeoData):
            raise RequestError("The geo data stream does not exist", 404)
        user = check_jwt_authentication(request)
        if stream.upload_user is None or stream.upload_user is not user:
            raise PermissionError(
                "The user is not the owner of the geo-data stream"
            )
        schema = StreamGeoDataAccessSchema()
        content = get_json_content()
        stream = schema.load(content, instance=stream)
        if not any([p.user is stream.upload_user for p in stream.permissions]):
            stream.permissions.append(
                BaseData.Permission(user=stream.upload_user)
            )
        stream.update()
        return dump_data(schema, access=stream)


class StreamDataListAPI(Resource):
    @token_required
    @with_response(
        status=200,
        schema=StreamGeoDataListResponseSchema,
        description="List of geodata streams the user has access to",
    )
    def get(self):
        """
        Get the list of all the geo-data stream for the user.

        :reqheader Authorization:
            JSON Web Token with Bearer scheme
            (``Authorization: Bearer <token>``).
        :status 403: No user logged in.

        .. :quickref: StreamGeoData; Get the list of accessible geodata streams
        """
        user = check_jwt_authentication(request)
        if user is None:
            raise PermissionError("No user logged in")
        stream_list = geo_data_stream_service.get_all_authorized_geo_data(user)
        return dump_data(StreamGeoDataSchema(many=True), stream=stream_list)


class FeatureAPI(Resource):
    @token_required
    @with_response(
        status=200,
        schema=FeatureResponseSchema,
        description="Geodata feature details",
    )
    def get(self, feature_id):
        """
        Return the details of the feature.

        :param int feature_id: The id of the feature.
        :reqheader Authorization:
            JSON Web Token with Bearer scheme
            (``Authorization: Bearer <token>``).
        :status 403:
            The user is not allowed to access the geo-data with this feature.
        :status 404: The geo-data feature doesn't exist.

        .. :quickref: Feature; Get the feature details
        """
        feature = Feature.get_by_id(feature_id)
        if feature is None:
            raise RequestError("The geo data feature does not exist", 404)
        geo_data_service.check_permission(feature.data)
        return dump_data(FeatureSchema(), feature=feature)


class FeatureInputsAPI(Resource):
    @token_required
    @with_response(
        status=200,
        schema=BaseDataFeaturesSchema,
        description="Data feature inputs",
    )
    def get(self, feature_id, data_id):
        """
        Return the inputs of the feature in specified data.

        :param int feature_id: The id of the feature.
        :param int data_id: The id of the input data.
        :reqheader Authorization:
            JSON Web Token with Bearer scheme
            (``Authorization: Bearer <token>``).
        :status 403:
            The user is not allowed to access the geo-data with this feature.
        :status 404: The geo-data feature or data doesn't exist.

        .. :quickref: Feature; Get the feature inputs from data.
        """
        feature = Feature.get_by_id(feature_id)
        if feature is None:
            raise RequestError("The geo data feature does not exist", 404)
        data = BaseData.get_by_id(data_id)
        if data is None:
            raise RequestError("The input data does not exist", 404)
        geo_data_service.check_permission(feature.data)
        features = feature.get_input_features(data_id)
        return jsonify({"features": [f.id for f in features]})


class FeatureOutputsAPI(Resource):
    @token_required
    @with_response(
        status=200,
        schema=BaseDataFeaturesSchema,
        description="Data feature outputs",
    )
    def get(self, feature_id, data_id):
        """
        Return the outputs of the feature in specified data.

        :param int feature_id: The id of the feature.
        :param int data_id: The id of the output data.
        :reqheader Authorization:
            JSON Web Token with Bearer scheme
            (``Authorization: Bearer <token>``).
        :status 403:
            The user is not allowed to access the geo-data with this feature.
        :status 404: The geo-data feature or data doesn't exist.

        .. :quickref: Feature; Get the feature outputs from data.
        """
        feature = Feature.get_by_id(feature_id)
        if feature is None:
            raise RequestError("The geo data feature does not exist", 404)
        data = BaseData.get_by_id(data_id)
        if data is None:
            raise RequestError("The output data does not exist", 404)
        geo_data_service.check_permission(feature.data)
        features = feature.get_output_features(data_id)
        return jsonify({"features": [f.id for f in features]})


class GeoDataVectorTileAPI(Resource):
    @token_required
    @with_response(
        status=200,
        description="Requested tile (might be empty)",
        content={
            "application/vnd.mapbox-vector-tile": {
                "type": "string",
                "format": "binary",
            }
        },
    )
    @with_query_arg(
        "property",
        str,
        required=False,
        description=(
            "The name of the property to return as the 'value' attribute. "
            "Default: None"
        ),
    )
    def get(self, geo_data_id, z, x, y):
        """
        Return a Mapbox Vector Tile of the geo-data.

        :param int geo_data_id: The id of the geo-data.
        :param int z: The tile zoom level.
        :param int x: The tile x index.
        :param int y: The tile y index.
        :query str property:
            The name of the property to return as the ``value`` attribute.
            None by default.
        :reqheader Authorization:
            JSON Web Token with Bearer scheme
            (``Authorization: Bearer <token>``).
        :status 403: The user is not allowed to access this geo-data.
        :status 404: The geo-data doesn't exist.

        .. :quickref: Geodata; Get a vector tile
        """
        geo_data_service.get_if_authorized(geo_data_id)
        parser = reqparse.RequestParser()
        parser.add_argument("property", type=str, location="args")
        args = parser.parse_args()
        property_arg = args.get("property", None)
        mvt = geo_data_service.get_vector_tile(
            geo_data_id, z, x, y, property_name=property_arg
        )
        return send_file(
            io.BytesIO(mvt), mimetype="application/vnd.mapbox-vector-tile"
        )
